from __future__ import absolute_import

import json
import logging
import sys
import os

import yaml

from logging import Formatter, StreamHandler, INFO
from threading import current_thread
from typing import Text, Dict

from appyratus.utils.time_utils import TimeUtils
from appyratus.utils.string_utils import StringUtils

from ravel.util.json_encoder import JsonEncoder


class LoggerInterface(object):
    def __init__(self, name, level=None, handlers=None):
        self._name = name
        self._level = level or INFO
        self._formatter = Formatter('%(message)s')

        self._logger = logging.getLogger(self._name)

        if self._level:
            self._logger.setLevel(self._level)
        else:
            self._level = self._logger.level

        if handlers:
            self._handlers = handlers
        else:
            stderr_handler = StreamHandler()
            stderr_handler.setLevel(self._level)
            self._handlers = [stderr_handler]

        for handler in self._handlers:
            if not handler.formatter:
                handler.setFormatter(self._formatter)
            self._logger.addHandler(handler)

    @property
    def logger(self):
        return self._logger

    def process_message(self, level: Text, message: Text, data: Dict) -> Text:
        return message

    def debug(self, message=None, data: Dict = None):
        self._logger.debug(self.process_message('debug', message, data))

    def info(self, message=None, data: Dict = None):
        self._logger.info(self.process_message('info', message, data))

    def warning(self, message=None, data: Dict = None):
        self._logger.warning(self.process_message('warning', message, data))

    def critical(self, message=None, data: Dict = None):
        self._logger.critical(self.process_message('critical', message, data))

    def error(self, message=None, data: Dict = None):
        self._logger.error(self.process_message('error', message, data))

    def exception(self, message=None, data: Dict = None):
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        data = (data if data is not None else{}).update({
            'exc_type': exc_type,
            'fname': fname,
            'lineno': exc_tb.tb_lineno
        })
        self._logger.exception(self.process_message('error', message, data))


class ConsoleLoggerInterface(LoggerInterface):
    def __init__(self, name, level=None, style=None):
        super().__init__(name, level=level)
        self._json = JsonEncoder()
        self._style = style or 'json'

    def process_message(self, level: Text, message: Text, data: Dict) -> Text:
        when = TimeUtils.utc_now().strftime('%m/%d/%Y %H:%M:%S')
        level = level.upper()[0]

        if data:
            data = self._json.decode(self._json.encode(data))
            if self._style == 'json':
                dumped_data = self._to_json(data)
            elif self._style == 'yaml':
                dumped_data = self._to_yaml(data)
            else:
                raise ValueError(f'unrcognized log style: {self.style}')
        else:
            dumped_data = None

        thread = StringUtils.dash(current_thread().name)

        display_string = (
            f'{when} ({level}) {self._name}'
            f' - {message}'
        )
        if dumped_data:
            display_string += f'\n\n{dumped_data}\n'
        return display_string

    def _to_json(self, data):
        return json.dumps(data, indent=2, sort_keys=True)

    def _to_yaml(self, data):
        lines = yaml.dump(data, default_flow_style=False).split('\n')
        return '\n'.join('  ' + line for line in lines)



from typing import Dict, List, Type, Set, Text

from ravel.util.misc_functions import is_sequence, get_class_name
from ravel.util.loggers import console

# XXX: This stuff needs some refactoring. Too hacky.


class BizBinding(object):

    def __init__(self, binder, resource_type, store_instance, store_bind_kwargs=None):
        self.binder = binder
        self.resource_type = resource_type
        self.store_instance = store_instance
        self.store_bind_kwargs = store_bind_kwargs or {}
        self._is_bound = False

    def __repr__(self):
        return (
            f'BizBinding({self.resource_type_name}, {self.store_class_name}, '
            f'bound={self.is_bound})'
        )

    def bind(self, binder=None, force=False):
        if self._is_bound:
            if not force:
                console.warning(
                    message=(
                        'resource store binding already set. '
                        'skipping repeated bind call'
                    ),
                    data={'resource_type': get_class_name(self.resource_type)}
                )
                return

        binder = binder or self.binder

        # associate a singleton Store instance with the res class.
        self.store_instance.bind(self.resource_type, **self.store_bind_kwargs)

        # first call bind on the Resource class itself
        self.resource_type.bind(binder)

        self._is_bound = True

    @property
    def is_bound(self):
        return self._is_bound

    @property
    def store_type(self):
        return self.store_instance.__class__

    @property
    def store_class_name(self):
        return get_class_name(self)

    @property
    def resource_type_name(self):
        return get_class_name(self.resource_type)


class ResourceBinder(object):
    """
    Stores and manages a global app, entailing which Resource class is
    associated with which Store class.
    """

    def __init__(self):
        self._bindings = {}
        self._named_store_types = {}
        self._named_resource_types = {}

    def __repr__(self):
        return f'{get_class_name(self)}()'

    @property
    def bindings(self) -> List['BizBinding']:
        return list(self._bindings.values())

    @property
    def resource_types(self) -> Dict[Text, 'Resource']:
        return self._named_resource_types

    @property
    def store_types(self) -> Dict[Text, 'Store']:
        return self._named_store_types

    def get_binding(self, resource_type):
        if isinstance(resource_type, type):
            resource_type = get_class_name(resource_type)
        return self._bindings.get(resource_type)

    def register(
        self,
        resource_type: Type['Resource'],
        store_type: Type['Store'],
        store_instance: 'Store' = None,
        store_bind_kwargs: Dict = None,
    ):
        store_class_name = get_class_name(store_type)
        if store_class_name not in self._named_store_types:
            store_type = type(store_class_name, (store_type, ), {})
            self._named_store_types[store_class_name] = store_type

        if store_instance is not None:
            assert isinstance(store_instance, store_type)
        else:
            store_instance = store_type()

        if resource_type is not None:
            resource_type_name = get_class_name(resource_type)
            resource_type.binder = self
            self._named_resource_types[resource_type_name] = resource_type
            self._bindings[resource_type_name] = binding = BizBinding(
                self,
                resource_type=resource_type,
                store_instance=store_instance,
                store_bind_kwargs=store_bind_kwargs,
            )
            return binding

        return None

    def bind(self, resource_types: Set[Type['Resource']] = None, rebind=False):
        if not resource_types:
            resource_types = [v.resource_type for v in self._bindings.values()]
        elif not is_sequence(resource_types):
            resource_types = [resource_types]
        for resource_type in resource_types:
            if not resource_type.ravel.is_abstract:
                resource_type.binder = self
                self.get_store_instance(resource_type, rebind=rebind)

    def get_store_instance(
        self,
        resource_type: Type['Resource'],
        bind=True,
        rebind=False,
    ) -> 'Store':
        if isinstance(resource_type, str):
            binding = self._bindings.get(resource_type)
        else:
            binding = self._bindings.get(get_class_name(resource_type))

        if binding is None:
            # lazily register a new binding
            base_store_type = resource_type.__store__()
            console.debug(
                f'calling {get_class_name(resource_type)}.__store__()'
            )
            binding = self.register(resource_type, base_store_type)

        # call bind only if it already hasn't been called
        if rebind or ((not binding.is_bound) and bind):
            console.debug(
                message=(
                    f'setting {get_class_name(binding.resource_type)}.ravel.store '
                    f'= {get_class_name(binding.store_instance)}()'
                )
            )
            binding.bind(binder=self)

        return binding.store_instance

    def get_store_type(self, store_class_name: Text) -> Type['Store']:
        return self._named_store_types.get(store_class_name)

    def is_registered(self, resource_type: Type['Resource']) -> bool:
        if isinstance(resource_type, str):
            return resource_type in self._bindings
        else:
            return get_class_name(resource_type) in self._bindings

    def is_bound(self, resource_type: Type['Resource']) -> bool:
        if isinstance(resource_type, str):
            return self._bindings[resource_type].is_bound
        else:
            return self._bindings[get_class_name(resource_type)].is_bound



import sys

from typing import Text, Dict, List

from appyratus.cli import CliProgram, PositionalArg, OptionalArg

from ravel.util.misc_functions import import_object


class ApplicationRouter(CliProgram):
    """
    # Application Router
    Program convenience for routing commands to multiple
    applications through a single command-line interface
    """

    def __init__(
        self,
        manifest: 'Manifest' = None,
        applications: Dict = None,
        default_app: Text = None,
        *args,
        **kwargs
    ):
        """
        # Init
        Do not merge unknown args into the args dict, as the app router
        only cares about the app field and nothing else.
        """
        self._manifest = manifest
        self._default_app = default_app
        self._apps = applications or {}

        # if applications were specified by dotted-path,
        # we perform the import here dynamically.
        for k, v in self._apps.items():
            if isinstance(v, str):
                self._apps[k] = import_object(v)
        super().__init__(merge_unknown=False, *args, **kwargs)

    def args(self):
        """
        # Args
        A list of arguments in use by the app router.  The first
        argument being the app that the CLI request will be
        routed to
        """
        app_names = ', '.join([r for r in self._apps.keys()])
        app_arg_args = {'name': 'app', 'usage': f'the app to utilize [{app_names}]'}
        if self._default_app:
            # default app is provided, so we will not require the app positional key
            app_arg = OptionalArg(**app_arg_args, default=self._default_app)
        else:
            app_arg = PositionalArg(**app_arg_args)
        return [app_arg]

    @staticmethod
    def perform(program):
        """
        # Perform routing
        Route to the app provided in the CLI's first argument
        """
        app_name = program.cli_args.app
        app = program.get_app(app_name)
        if not app:
            raise Exception(f'Unable to locate app "{app_name}"')
        app_method = getattr(program, f'run_{app_name}', None)
        if callable(app_method):
            return app_method(app)
        else:
            program.app_lifecycle(app=app)

    def get_app(self, name: Text):
        """
        # Get app
        First by applications dictionary (provided when initialized)
        And if not there, then an attribute on this your router class
        """
        if not self._apps:
            app_dict = {}
        else:
            app_dict = self._apps.get(name)
        app_attr = getattr(self, name, None)
        if app_dict:
            return app_dict
        elif app_attr:
            return app_attr()

    def app_lifecycle(
        self,
        app: 'Application',
        manifest: Text = None,
        bootstrap_kwargs: Dict = None,
        start_kwargs: Dict = None
    ):
        """
        Application lifecycle
        Perform necessary bootstrapping with manifest and then fire it up
        """
        app.bootstrap(manifest=manifest or self._manifest, **(bootstrap_kwargs or {}))
        app.start(**(start_kwargs or {}))
        return app

    def run_cli(self, cli_app: 'CliApplication'):
        """
        # Run Cli
        A special implementation for the CLI app to provide this program's
        unknown CLI args to the CLI app program.
        """
        return self.app_lifecycle(
            app=cli_app, bootstrap_kwargs={'cli_args': self._unknown_cli_args}
        )



from typing import Dict, Tuple, Text, Type, Callable
from celery.signals import worker_process_init

import kombu.serialization

from mock import MagicMock
from celery import Celery

from appyratus.utils.string_utils import StringUtils
from appyratus.memoize import memoized_property

from ravel.util import get_class_name, is_resource, is_batch
from ravel.util.scanner import Scanner
from ravel.util.loggers import console
from ravel.manifest import Manifest
from ravel.app.base import Application, Action

from .celery_task import CeleryTask

DEFAULT_BROKER = 'amqp://'


class CeleryService(Application):

    def __init__(self, uses_client=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._client = CeleryClient(self) if uses_client else None
        self._celery = None

    @property
    def action_type(self) -> Type['CeleryTask']:
        return CeleryTask

    @property
    def celery_config(self) -> Dict:
        return self.local.manifest.data.get('celery', {})

    @property
    def client(self) -> 'CeleryClient':
        if self._client is None:
            self._client = CeleryClient(self)
        return self._client

    @property
    def celery(self) -> Celery:
        return self._celery

    def on_bootstrap(self, *args, **kwargs):
        self.initialize_celery()

    def on_start(self) -> Celery:
        return self.celery

    def initialize_celery(self):
        self._init_celery_app()
        self._init_celery_json_serializer()
        for action in self.actions.values():
            action.register_with_celery()

    def _init_celery_app(self):
        name = self.local.manifest.package
        broker = self.celery_config.setdefault('broker', DEFAULT_BROKER)
        backend = self.celery_config.get('result_backend')
        self._celery = Celery(name, broker=broker, backend=backend)
        self._celery.conf.update(self.celery_config)

        @worker_process_init.connect
        def bootstrap_celery_worker_process(*args, **kwargs):
            console.info('bootstrapping celery worker process')
            self.bootstrap(self.local.manifest)

    def _init_celery_json_serializer(self):
        serializer_name = 'json'
        kombu.serialization.register(
            serializer_name,
            self.json.encode,
            self.json.decode,
            content_type='application/json',
            content_encoding='utf-8',
        )
        self._celery.task_serializer = serializer_name
        self._celery.conf.update(
            task_serializer=serializer_name,
            accept_content=[serializer_name],
            result_serializer=serializer_name,
        )


class CeleryClient(object):

    def __init__(self, app):
        self.app = app
        self.scanner = Scanner()
        self.methods = {}

    def __getattr__(self, method: Text) -> Callable:
        method = self.methods.get(method)
        if method is None:
            console.error(
                message=f'celery client does not recognize task name',
                data={'task': method}
            )
            raise ValueError(method)
        return method

    def bootstrap(self, manifest: 'Manifest', mocked=False) -> 'CeleryClient':
        # discover actions registered with app via decorator
        manifest = Manifest(manifest)
        self.scanner.scan(manifest.package)

        if not mocked:
            # trigger creation of the native Celery task objects
            self.app.manifest = manifest
            self.app.initialize_celery()
            console.info(
                message=f'initializing Celery client',
                data={
                    'broker': self.app.celery.conf.broker_url,
                    'tasks': sorted(self.app.actions.keys()),
                }
            )
            for action in self.app.actions.values():
                action.bootstrap()
                self.methods[action.name] = action
        else:
            # if "mocked", do not bother doing anything but
            # creating mock client methods
            for action in self.app.actions.values():
                self.methods[action.name] = MagicMock()

        return self



import traceback

from typing import Text, Dict, Type, Union, Set

from ravel.util.loggers import console
from ravel.util.scanner import Scanner


class ManifestScanner(Scanner):
    """
    The manifest's type scanner is used to detect Store and Resource classes.
    """

    def __init__(self, manifest: 'Manifest', *args, **kwargs):
        super().__init__(*args, **kwargs)

        from ravel.store.base.store import Store
        from ravel.resource import Resource

        self.manifest = manifest
        self.base_store_class = Store
        self.base_resource_class = Resource

        self.context.store_classes = {}
        self.context.resource_classes = {}

    def predicate(self, value) -> bool:
        """
        Predicate should evaluate True for any object we want this scanner to
        match and pass on to the self.on_match callback.
        """
        def is_resource_or_store_class(value) -> bool:
            return (
                isinstance(value, type)
                and issubclass(value, (
                    self.base_resource_class,
                    self.base_store_class
                ))
            )

        return (
            is_resource_or_store_class(value)
        )

    def on_match(self, name: Text, value, context):
        """
        Add the class object to the appripriate container.
        """
        if isinstance(value, type):
            if issubclass(value, self.base_resource_class):
                if not value.ravel.is_abstract:
                    context.resource_classes[name] = value
            elif issubclass(value, self.base_store_class):
                context.store_classes[name] = value

    def on_import_error(self, exc: Exception, module_name: Text, context):
        """
        If the scanner fails to import a module while it walks the file
        system, it comes here to handle and report the problem.
        """
        exc_str = traceback.format_exc()
        console.error(
            message=f'manifest could not scan module: {module_name}',
            data={'traceback': exc_str.split('\n')}
        )

    def on_match_error(self, exc: Exception, module, context, name, value):
        """
        If there's a problem in self.on_match, we come here.
        """
        exc_str = traceback.format_exc()
        console.error(
            message=f'error while scanning {name} ({type(value)})',
            data={'traceback': exc_str.split('\n')}
        )




from ravel.util.misc_functions import (
    is_sequence,
)


class DirtyObject(object):
    def __init__(self, *args, **kwargs):
        self._dirty = set()

    @property
    def dirty(self):
        return self._dirty

    def clean(self, keys=None):
        if keys:
            for k in keys:
                self.dirty.discard(k)
        else:
            self.dirty.clear()

    def mark(self, keys=None):
        if keys:
            for k in keys:
                self.dirty.add(k)
        else:
            self.dirty.clear()


class DirtyDict(DirtyObject, dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        DirtyObject.__init__(self)
        self.dirty.update(self.keys())

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.dirty.add(key)

    def __delitem__(self, key):
        super().__delitem__(key)
        self.dirty.discard(key)

    def mark(self, keys=None):
        if not is_sequence(keys):
            keys = {keys}
        keys = {
            k for k in keys if k in self
        }
        super().mark(keys)

    def pop(self, key, default=None):
        value = super().pop(key, default)
        self.dirty.discard(key)
        return value

    def update(self, data):
        super().update(data)
        self.dirty.update(data.keys())

    def setdefault(self, key, default):
        if key not in self:
            self.dirty.add(key)
        return super().setdefault(key, default)


class DirtyList(DirtyObject, list):
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        DirtObject.__init__(self)
        self.dirty.update(range(len(self)))

    def __setitem__(self, index, value):
        super().__setitem__(index, value)
        self.dirty.add(index)

    def __delitem__(self, index):
        super().__delitem__(index)
        self.dirty.add(index)

    def append(self, value):
        super().append(value)
        self.dirty.add(len(self) - 1)

    def pop(self):
        value = super().pop()
        self.dirty.remove(len(self) - 1)
        return value



from random import randint
from typing import Text, Tuple, List, Set, Dict, Type, Union, Callable
from collections import defaultdict, deque
from itertools import islice

from appyratus.utils.dict_utils import DictObject
from BTrees.OOBTree import BTree

from ravel.util.misc_functions import (
    get_class_name,
    flatten_sequence,
    union,
)
from ravel.query.predicate import (
    Predicate,
    ConditionalPredicate,
    BooleanPredicate,
)
from ravel.constants import IS_BATCH, OP_CODE
from ravel.schema import fields
from ravel.util.loggers import console
from ravel.util import is_batch, is_resource
from ravel.entity import Entity
from ravel.query.order_by import OrderBy


class Batch(Entity):
    """
    A Batch is a collection of Resources. At runtime, Ravel creates a new Batch
    subclass for each Resource type in the app. For example, a User resource
    will have User.Batch. An Account resource will have Account.Batch. Each
    Batch subclass has non-scalar versions of the ResolverProperties owned by
    the corresponding Resource type.

    ## Accessing Resource State in Batch
    Resource and Batch types both implement the same property-based system for
    accessing and writing state, allowing both `user.name` to access the name of
    a single User and `users.name`, where `users` is a User.Batch instance, to
    access a list of all names in the batch. For example:

    ```python
    users = User.Batch([User(name='Augustus'), User(name='Caligula')])
    assert users.name == ['Augustus', 'Caligula']
    ```

    ## CRUD
    Batches implement the same CRUD interface as Resource types, performing the
    batch version of each operation. For example, `users.create()` will call
    `User.create_many(users)` under the hood.

    ## Filtering
    It is possible to filter batches with a "where" method. For this to work,
    a batch must be indexed. Indexing is enabled by default but can be toggled
    through a constructor argument. Filtering is O(log N). The "where" method
    returns a new batch containing the filtered results. This looks like:

    ```python
    filered_users = users.where(User.name == 'Caligula')
    ```

    ## Random Access
    It is possible to access specific resources by indexing a batch like an
    array. For example `users[1:]`. Note that `users["name"]` will not work; for
    that, do `users.name`.
    """

    ravel = DictObject()
    ravel.owner = None
    ravel.properties = {}
    ravel.indexed_field_types = (
        fields.String, fields.Int, fields.Id,
        fields.Bool, fields.Float
    )

    def __init__(self, resources: List = None, indexed=True):
        self.internal = DictObject()
        self.internal.resources = deque(resources or [])
        self.internal.indexed = indexed
        self.internal.indexes = defaultdict(BTree)
        if indexed:
            self.internal.indexes.update({
                k: BTree()
                for k, field in self.ravel.owner.Schema.fields.items()
                if isinstance(field, self.ravel.indexed_field_types)
            })

    def __len__(self):
        return len(self.internal.resources)

    def __bool__(self):
        return bool(self.internal.resources)

    def __iter__(self):
        return iter(self.internal.resources)

    def __getitem__(self, index):
        """
        Return a single resource or a new batch, for the given positional index
        or slice.
        """
        if isinstance(index, slice):
            return type(self)(
                islice(
                    self.internal.resources,
                    index.start, index.stop, index.step
                )
            )
        else:
            assert isinstance(index, int)
            return self.internal.resources[index]

    def __setitem__(self, index: int, resource: 'Resource'):
        """
        Write a resource to a specified index. The resource being written to the
        batch must belong to the same type as all other elements. That type is
        stored in `Batch.ravel.owner`.
        """
        owner = self.ravel.owner
        if owner and not isinstance(resource, owner):
            raise ValueError(f'wrong Resource type')
        self.internal.resources[index] = value

    def __repr__(self):
        """
        Show the name of the batch, its owner resource type, and its size.
        """
        return (
            f'{get_class_name(self.ravel.owner)}.Batch('
            f'size={len(self)})'
        )

    def __add__(self, other):
        """
        Create and return a copy, containing the concatenated data lists.
        """
        clone = type(self)(self.internal.resources, indexed=self.indexed)

        if isinstance(other, (list, tuple, set)):
            clone.internal.resources.extend(other)
        elif isinstance(other, Batch):
            assert other.ravel.owner is self.ravel.owner
            clone.internal.resources.extend(other.internal.resources)
        else:
            raise ValueError(str(other))

        return clone

    @classmethod
    def factory(cls, owner: Type['Resource'], type_name=None):
        """
        Thie factory method returns new Batch subtypes and is used internally at
        runtime by the Resource metaclass. For example, if you have a Resource
        type called `User`, then this method is invoked to create `User.Batch`.
        """
        type_name = type_name or 'Batch'  # name of new class

        # start with inherited ravel object
        ravel = DictObject()

        ravel.owner = owner
        ravel.indexed_field_types = (
            fields.String, fields.Int, fields.Id,
            fields.Bool, fields.Float
        )
        ravel.properties = {}

        for k, resolver in owner.ravel.resolvers.items():
            ravel.properties[k] = BatchResolverProperty(resolver)

        derived_batch_type = type(type_name, (cls, ), dict(
            ravel=ravel, **ravel.properties
        ))

        # IS_BATCH is a type flag used in is_batch utility methods
        setattr(derived_batch_type, IS_BATCH, True)

        return derived_batch_type

    @classmethod
    def generate(
        cls,
        resolvers: Set[Text] = None,
        values: Dict = None,
        count: int = None,
    ):
        """
        Generate a Batch of fixtures, generating random values for the
        specified resolvers.

        ## Arguments
        - `resolvers`: list of resolver names for which to generate values.
        - `values`: dict of static values to use.
        - `count`: number of resources to generate. defaults between [1, 10].
        """
        owner = cls.ravel.owner

        # if no count, randomize between [1, 10]
        count = max(count, 1) if count is not None else randint(1, 10)

        if owner is None:
            # this batch isn't associated with any Resource type
            # and therefore we don't know what to generate.
            raise Exception('unbound Batch type')
        if not resolvers:
            resolvers = set(owner.ravel.resolvers.fields.keys())
        else:
            resolvers = set(resolvers)

        # create and return the new batch
        return cls(
            owner.generate(resolvers, values=values)
            for _ in range(count)
        )

    def foreach(self, callback: Callable) -> 'Batch':
        for i, x in enumerate(self.internal.resources):
            callback(i, x)
        return self

    def sort(self, order_by) -> 'Batch':
        self.internal.resources = OrderBy.sort(
            self.internal.resources, order_by
        )
        return self

    def set(
        self,
        other: Union[Dict, 'Resource'] = None,
        **values
    ) -> 'Batch':
        return self.merge(other=other, **values)

    def merge(
        self,
        other: Union[Dict, 'Resource'] = None,
        **values
    ) -> 'Batch':
        """
        Batch merge the given dict or resource into all resources contained in
        the batch.
        """
        for resource in self.internal.resources:
            resource.merge(other)
            resource.merge(values)
        return self

    def insert(self, index, resource):
        """
        Insert a resource at the specified index.
        """
        self.internal.resources.insert(index, resource)
        if self.internal.indexed:
            self._update_indexes(resource)
        return self

    def remove(self, resource):
        """
        Remove a resource at the specified index.
        """
        try:
            self.internal.resources.remove(resource)
            self._prune_indexes(resource)
        except ValueError:
            raise
        return self

    def append(self, resource):
        """
        Add a resource at the end of the batch.
        """
        self.insert(-1, resource)
        return self

    def appendleft(self, resource):
        """
        Add a resource at the front of the batch (index: 0)
        """
        self.insert(0, resource)
        return self

    def extend(self, resources):
        """
        Insert a collection of resource at the end of the batch.
        """
        self.internal.resources.extend(resources)
        for resource in resources:
            self._update_indexes(resource)

    def extendleft(self, resources):
        """
        Insert a collection of resource at the front of the batch.
        """
        self.internal.resources.extendleft(resources)
        for resource in resources:
            self._update_indexes(resource)

    def pop(self, index=-1):
        """
        Remove and return the resource at the end of the batch.
        """
        resource = self.internal.resources.pop(index)
        self._prune_indexes(resource)
        return resource

    def popleft(self, index=-1):
        """
        Remove and return the resource at the front of the batch.
        """
        resource = self.internal.resources.popleft(index)
        self._prune_indexes(resource)

    def rotate(self, n=1):
        """
        Shift ever element to the right so that the element at index -1 wraps
        around, back to index 0. The direction can be reversed using a negative
        value of `n`.
        """
        self.internal.resources.rotate(n)
        return self

    def where(self, *predicates: Tuple['Predicate'], indexed=False) -> 'Batch':
        """
        Return a new batch, containing the resources whose attributes match the
        given predicates, like `User.created_at >= cutoff_date`, for instance.
        The new batch is *not* indexed by default, so make sure to set
        `indexed=True` if you intend to filter the filtered batch further.
        """
        predicate = Predicate.reduce_and(flatten_sequence(predicates))
        resources = self._apply_indexes(predicate)
        return type(self)(resources, indexed=indexed)

    def create(self):
        """
        Insert all resources into the store.
        """
        if self:
            self.ravel.owner.create_many(self.internal.resources)
        return self

    def update(self, state: Dict = None, **more_state):
        """
        Apply an update to all resources in the batch, writing it to the store.
        """
        if self:
            self.ravel.owner.update_many(self, data=state, **more_state)
        return self

    def delete(self):
        """
        Delete all resources in the batch fro mthe store.
        """
        if self:
            self.ravel.owner.delete_many(self.internal.resources)
        return self

    def save(self, resolvers: Union[Text, Set[Text]] = None):
        """
        Save all resources in the batch, effectively creating some and updating
        others.
        """
        if not self:
            return self

        # batch save resource resolver targets
        if resolvers is not None:
            owner_resource_type = self.ravel.owner
            fields_to_save = set()

            if isinstance(resolvers, str):
                resolvers = {resolvers}
            elif not isinstance(resolvers, set):
                resolvers = set(resolvers)

            for name in resolvers:
                resolver = owner_resource_type.ravel.resolvers[name]
                if name not in owner_resource_type.ravel.schema.fields:
                    visited_ids = set()
                    unique_targets = resolver.target.Batch()
                    if resolver.many:
                        for batch in getattr(self, name):
                            if batch:
                                unique_targets.extend(batch)
                    else:
                        for resource in getattr(self, name):
                            if resource:
                                unique_targets.append(resource)
                    if unique_targets:
                        unique_targets.save()
                else:
                    fields_to_save.add(name)
        else:
            fields_to_save = None

        # now save fields
        self.ravel.owner.save_many(
            self.internal.resources,
            resolvers=fields_to_save,
        )
        return self

    def clean(self, resolvers: Set[Text] = None):
        """
        Mark all resources in the batch as clean, meaning that no state elements
        are considered dirty and in need up saving to the store.
        """
        for resource in self.internal.resources:
            # TODO: renamed fields kwarg to resolvers
            resource.clean(fields=resolvers)
        return self

    def mark(self, resolvers: Set[Text] = None):
        """
        Mark the given resolvers as dirty, for all resources in the batch.
        """
        for resource in self.internal.resources:
            resource.mark(fields=resolvers)
        return self

    def dump(
        self,
        resolvers: Set[Text] = None,
        style: 'DumpStyle' = None,
    ) -> List[Dict]:
        """
        Return a list with the dump of each resource in the batch.
        """
        return [
            resource.dump(resolvers=resolvers, style=style)
            for resource in self.internal.resources
        ]

    def resolve(self, resolvers: Union[Text, Set[Text]] = None) -> 'Batch':
        """
        Execute each of the resolvers, specified by name, storing the results in
        `self.internal.state`.
        """
        if self._id is None:
            return self

        if isinstance(resolvers, str):
            resolvers = {resolvers}
        elif not resolvers:
            resolvers = set(self.ravel.owner.ravel.schema.fields.keys())

        # execute all requested resolvers
        for k in resolvers:
            resolver = self.ravel.owner.ravel.resolvers.get(k)
            if resolver is not None:
                resolver.resolve_batch(self)

        # clean the resolved values so they arent't accidently saved on
        # update/create, as we just fetched them from the store.
        self.clean(resolvers)

        return self

    def unload(self, resolvers: Set[Text] = None) -> 'Batch':
        """
        Clear resolved state data from all resources in the batch.
        """
        if not resolvers:
            resolvers = set(self.owner.Schema.fields.keys())
        resolvers.discard(ID)
        resolvers.discard(REV)
        for resource in self.internal.resources:
            resource.unload(resolvers)
        return self

    def _update_indexes(self, resource):
        """
        Insert a Resource from the index B-trees.
        """
        for k, index in self.internal.indexes.items():
            if k in resource.internal.state:
                value = resource[k]
                if value not in index:
                    index[value] = set()
                index[value].add(resource)

    def _prune_indexes(self, resource):
        """
        Remove a Resource from the index B-trees.
        """
        for k, index in self.internal.indexes.items():
            if k in resource.internal.state:
                value = resource[k]
                if value in index:
                    index[value].remove(resource)

    def _apply_indexes(self, predicate):
        if predicate is None:
            return self.records.keys()

        op = predicate.op
        empty = set()
        _ids = set()

        if isinstance(predicate, ConditionalPredicate):
            k = predicate.field.source
            v = predicate.value
            indexes = self.internal.indexes
            index = indexes[k]

            if op == OP_CODE.EQ:
                _ids = indexes[k].get(v, empty)
            elif op == OP_CODE.NEQ:
                _ids = union([
                    _id_set for v_idx, _id_set in index.items()
                    if v_idx != v
                ])
            elif op == OP_CODE.INCLUDING:
                v = v if isinstance(v, set) else set(v)
                _ids = union([index.get(k_idx, empty) for k_idx in v])
            elif op == OP_CODE.EXCLUDING:
                v = v if isinstance(v, set) else set(v)
                _ids = union([
                    _id_set for v_idx, _id_set in index.items()
                    if v_idx not in v
                ])
            else:
                keys = np.array(index.keys(), dtype=object)
                if op == OP_CODE.GEQ:
                    offset = bisect.bisect_left(keys, v)
                    interval = slice(offset, None, 1)
                elif op == OP_CODE.GT:
                    offset = bisect.bisect(keys, v)
                    interval = slice(offset, None, 1)
                elif op == OP_CODE.LT:
                    offset = bisect.bisect_left(keys, v)
                    interval = slice(0, offset, 1)
                elif op == OP_CODE.LEQ:
                    offset = bisect.bisect(keys, v)
                    interval = slice(0, offset, 1)
                else:
                    # XXX: raise StoreError
                    raise Exception('unrecognized op')
                _ids = union([
                    index[k] for k in keys[interval]
                    if k is not None
                ])
        elif isinstance(predicate, BooleanPredicate):
            lhs = predicate.lhs
            rhs = predicate.rhs
            if op == OP_CODE.AND:
                lhs_result = self._apply_indexes(lhs)
                if lhs_result:
                    rhs_result = self._apply_indexes(rhs)
                    _ids = set.intersection(lhs_result, rhs_result)
            elif op == OP_CODE.OR:
                lhs_result = self._apply_indexes(lhs)
                rhs_result = self._apply_indexes(rhs)
                _ids = set.union(lhs_result, rhs_result)
            else:
                # XXX: raise StoreError
                raise Exception('unrecognized boolean predicate')

        return _ids


class BatchResolverProperty(property):
    def __init__(self, resolver):
        self.resolver = resolver
        super().__init__(
            fget=self.fget,
            fset=self.fset,
            fdel=self.fdel,
        )

    def fget(self, batch: 'Batch'):
        key = self.resolver.name
        return [
            getattr(resource, key, None)
            for resource in batch.internal.resources
        ]

    def fset(self, batch: 'Batch', value):
        key = self.resolver.name
        for resource in batch.internal.resources:
            setattr(resource, key, value)

    def fdel(self, batch: 'Batch'):
        key = self.resolver.name
        for resource in batch.internal.resources:
            delattr(resource, key)



from typing import Dict, Tuple, Text, Type

from celery import Celery

from appyratus.utils.string_utils import StringUtils

from ravel.util import get_class_name, is_resource, is_batch
from ravel.app.base import Application, Action
from ravel.api import Api

DEFAULT_BROKER = 'amqp://'


class CeleryTask(Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._celery_task = None  # <- set in CeleryTaskManager.on_bootstrap

    @property
    def celery_task(self):
        return self._celery_task

    def s(self, *args, **kwargs):
        return self._celery_task.s(*args, **kwargs)

    def register_with_celery(self):
        # we must define a new function as the task target because
        # celery expects a __name__ attribute, which an Action object
        # itself does not have.
        func = lambda *args, **kwargs: self(*args, **kwargs)
        func.__name__ = f'{self.name}_task_function'

        task = self.app.celery.task(**self.decorator.kwargs)
        self._celery_task = task(func)

    def delay(self, *args, **kwargs):
        args, kwargs = self._prepare_celery_task_arguments(args, kwargs)
        return self._celery_task.delay(*args, **kwargs)

    def _prepare_celery_task_arguments(self, args, kwargs):
        args = args or tuple()
        kwargs = kwargs or {}

        def process_value(obj):
            if is_resource(obj):
                if obj._id is not None:
                    return obj._id
                else:
                    return obj.internal.state
            elif is_batch(obj):
                raise NotImplemented('TODO: implement support for batch objects')
            else:
                return obj

        if args and isinstance(args[0], Api):
            # ignore the "self" argument of the Api instance
            # whose method is registered as this action's callable.
            args = args[1:]

        new_args = [process_value(x) for x in args]
        new_kwargs = {k: process_value(v) for k, v in kwargs.items()}
        return (new_args, new_kwargs)



from ravel.schema import fields
from ravel.schema import Schema, fields, Id
from ravel.app.template import TemplateRenderer, JinjaTemplateRenderer
from ravel.manifest.manifest import Manifest
from ravel.resource import Resource, ResourceMeta
from ravel.logging import ConsoleLoggerInterface
from ravel.store.base.store import Store
from ravel.api import Api
from ravel.app.base import Application, ActionDecorator, Action
from ravel.app.middleware import Middleware
from ravel.resource import Resource
from ravel.entity import Entity
from ravel.batch import Batch
from ravel.util import is_resource, is_batch, is_resource_type, is_batch_type
from ravel.batch import Batch
from ravel.query.query import Query
from ravel.query.order_by import OrderBy
from ravel.query.request import Request
from ravel.query.predicate import (Predicate, ConditionalPredicate, BooleanPredicate)
from ravel.resolver.resolver import Resolver
from ravel.resolver.resolver_manager import ResolverManager
from ravel.resolver.resolver_decorator import ResolverDecorator
from ravel.resolver.resolver_property import ResolverProperty
from ravel.resolver.resolvers.loader import Loader, LoaderProperty
from ravel.resolver.resolvers.relationship import Relationship
from ravel.resolver.decorators import (resolver, relationship, view, nested, field)



from typing import Text, Type, Union, Callable

from ravel.util.misc_functions import get_callable_name

class ResolverDecorator(object):
    """
    Instances of ResolverDecorator are intended to register Resource instance
    methods as callback function arguments to a Resolver, which is constructed
    by the Resource metaclass.
    """

    def __init__(
        self,
        resolver_type: Type['Resolver'],
        *args,
        **kwargs
    ):
        self.resolver_type = resolver_type
        self.args = args
        self.kwargs = kwargs

    def __call__(self, func: Callable) -> 'ResolverDecorator':
        """
        The decorator's main purpose is to register a an instance method
        as a resolver's on_resolve callback, which is the main function of a
        Resolver.
        """
        self.kwargs['on_resolve'] = func
        self.kwargs.setdefault('name', get_callable_name(func))
        return self

    def __getattr__(self, key):
        """
        This allows the decorator object to have attributes that are themselves
        decorators, for the purpose of adding the decorated methods as
        kwargs to the resolver, with usage like:

        ```python

        @resolver(Account)
        def account(self, request):
            return Query(request=request).execute(first=True)

        @account.on_get
        def log_account_access(self, account):
            print(f'{account._id} accessed')
        ```
        """
        def insert(func):
            func.__name__ = f'{func.__name__}_{key}'
            self.kwargs[key] = func
            return self

        return insert

    def build_property(
        self, owner: Type['Resource'], name: Text
    ) -> 'ResolverProperty':
        """
        Return a new DecoratorProperty using the args and kwargs collected in
        this ResolverDecorator.
        """
        self.kwargs.update({'owner': owner, 'name': name})
        return self.resolver_type.build_property(
            decorator=self, args=self.args, kwargs=self.kwargs
        )



from typing import List, Dict, Text, Type, Set, Tuple

from appyratus.env import Environment
from ravel.app.middleware import Middleware, MiddlewareError
from ravel.util.misc_functions import get_class_name
from ravel.util.loggers import console

from .store import SqlalchemyStore

ADD_POST_COMMIT_HOOK_METHOD = 'add_post_commit_hook'
POST_COMMIT_HOOKS = 'ManageSqlalchemyTransaction_post_commit_hooks'

class ManageSqlalchemyTransaction(Middleware):
    """
    Manages a Sqlalchemy database transaction that encompasses the execution of
    an Action.
    """

    def __init__(self, store_class_name: Text = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.env = Environment()
        self.store_class_name = store_class_name or 'SqlalchemyStore'

    def on_bootstrap(self):
        """
        Aqcuire the Store subclass object from self.store_class_name.
        """
        store_types = self.app.manifest.store_classes
        self.store_type = store_types.get(self.store_class_name)
        if self.store_type is None:
            raise MiddlewareError(
                self, f'{self.store_class_name} class not found'
            )

    def pre_request(
        self,
        action: 'Action',
        request: 'Request',
        raw_args: Tuple,
        raw_kwargs: Dict
    ):
        """
        Get a connection from Sqlalchemy's connection pool and begin a
        transaction.
        """
        hooks = []

        # add a dynamic add_post_commit_hook method to request.context
        # for use in Actions that register post-commit hooks.
        request.context[POST_COMMIT_HOOKS] = hooks
        request.context[ADD_POST_COMMIT_HOOK_METHOD] = func = (
            lambda hook, args=(), kwargs={}: hooks.append(
                [hook, args, kwargs]
            )
        )
        # add post-commit hook storage containers to
        # application thread-local state
        setattr(self.app.local, POST_COMMIT_HOOKS, hooks)
        setattr(self.app.local, ADD_POST_COMMIT_HOOK_METHOD, func)

        self.store_type.begin()

    def post_request(
        self,
        action: 'Action',
        request: 'Request',
        result,
    ):
        """
        Commit or rollback the tranaction.
        """
        self.store_type.commit(rollback=True)

        # execute post-commit hooks in background processes
        for hook, args, kwargs in request.context[POST_COMMIT_HOOKS]:
            self.app.spawn(
                hook, args=args, kwargs=kwargs, multiprocessing=True
            )

    def post_bad_request(
        self,
        action: 'Action',
        request: 'Request',
        exc: Exception,
    ):
        """
        Rollback a failed transaction.
        """
        console.info(f'rolling back sqlalchemy transaction')
        try:
            self.store_type.rollback()
        finally:
            self.store_type.close()



from typing import Text, Dict

from appyratus.utils.dict_utils import DictObject

from ravel.util import is_resource, is_resource_type
from ravel.util.misc_functions import get_class_name, flatten_sequence

from .parameters import ParameterAssignment


class Request(object):

    # Query is set at runtime as a side-effect of
    # importing the Query module.
    Query = None


    def __init__(
        self,
        resolver: 'Resolver',
        query: 'Query' = None,
        parent: 'Request' = None,
    ):
        self.resolver = resolver
        self.parameters = DictObject({'select': []})
        self.query = query
        self.result = None
        self.parent = parent

    def __repr__(self):
        return (
            f'{get_class_name(self)}('
            f'target={get_class_name(self.resolver.owner)}.'
            f'{self.resolver.name}'
            f')'
        )

    def __getattr__(self, name: Text) -> 'ParameterAssignment':
        return ParameterAssignment(self, name)

    def select(self, *args) -> 'Request':
        from ravel.resolver.resolvers.loader import LoaderProperty
        from ravel.resolver.resolver_property import ResolverProperty

        args = flatten_sequence(args)

        for obj in args:
            if isinstance(obj, str):
                # if obj is str, replace it with the corresponding resolver
                # property from the target Resource class.
                _obj = getattr(self.resolver.owner, obj, None)
                if _obj is None:
                    raise ValueError(f'unknown resolver: {obj}')
                obj = _obj
            elif is_resource(obj):
                self.select(obj.internal.state.keys())
                continue
            elif is_resource_type(obj):
                self.select(obj.ravel.resolvers.keys())
                continue

            # build a resolver request
            request = None
            if isinstance(obj, LoaderProperty) and (obj.decorator is None):
                resolver_property = obj
                request = Request(resolver_property.resolver)
            elif isinstance(obj, ResolverProperty):
                resolver_property = obj
                request = Request(resolver_property.resolver)
            elif isinstance(obj, Request):
                request = obj

            if request is not None:
                self.parameters.select.append(request)
            elif obj is not None:
                self.parameters.select.append(obj)

        return self


from typing import Text
from uuid import UUID

from ravel.entity import Entity
from ravel.constants import (
    IS_RESOURCE,
    IS_BATCH,
    ID,
)


def is_resource(obj):
    """
    Return True if obj is an instance of Resource.
    """
    return (
        isinstance(obj, Entity)
        and getattr(obj, IS_RESOURCE, False)
    )


def is_batch(obj) -> bool:
    """
    Return True if obj is an instance of Resource.
    """
    return (
        isinstance(obj, Entity)
        and getattr(obj, IS_BATCH, False)
    )


def is_resource_type(obj):
    return (
        isinstance(obj, type)
        and getattr(obj, IS_RESOURCE, False)
    )


def is_batch_type(obj):
    return (
        isinstance(obj, type)
        and getattr(obj, IS_BATCH, False)
    )


def repr_res_id(resource: 'Resource') -> Text:
    """
    Return a string version of a Resource ID.
    """
    # TODO: rename this function
    if resource is None:
        return 'None'

    _id = resource[ID]

    if _id is None:
        return '?'
    elif isinstance(_id, str):
        return _id
    elif isinstance(_id, UUID):
        return _id.hex
    else:
        return repr(_id)



from appyratus.env import Environment
from appyratus.enum import Enum


ENV = Environment()

# logging:
CONSOLE_LOG_LEVEL = ENV.get('RAVEL_CONSOLE_LOG_LEVEL', 'INFO')


# field name constants:
ID = '_id'
REV = '_rev'


# entity class annotations used for type checking:
IS_RESOURCE = '_ravel_is_resource'
IS_BATCH = '_ravel_is_batch'


# op codes for Predicate objects
OP_CODE = Enum(
    EQ='eq',
    NEQ='neq',
    GT='gt',
    LT='lt',
    GEQ='geq',
    LEQ='leq',
    INCLUDING='in',
    EXCLUDING='ex',
    AND='and',
    OR='or',
)



from typing import Text, Dict

from ravel.schema import Schema, fields


class Bootstrap:

    class Schema(Schema):
        store = fields.String()
        default = fields.Bool(default=False)
        params = fields.Dict(default={})

    def __init__(
        self,
        store_class_name: Text,
        bootstrap_params: Dict = None,
        is_default: bool = False,
    ):
        self.store_class_name = store_class_name
        self.bootstrap_params = bootstrap_params or {}
        self.is_default = is_default


from typing import Type

from appyratus.utils.time_utils import TimeUtils

from ravel import Resource, Id, relationship, fields
from ravel.query.predicate import Predicate


class Session(Resource):
    is_active = fields.Bool(nullable=False, default=lambda: True)
    logged_out_at = fields.DateTime(nullable=True, default=lambda: None)

    @classmethod
    def __abstract__(cls):
        return True

    def logout(self):
        if self.is_active:
            self.update(logged_out_at=TimeUtils.utc_now(), is_active=False)



import re

from typing import Dict, Set, Text, List, Type, Tuple, Union

from ravel.query.predicate import Predicate
from ravel.query.order_by import OrderBy


class GraphqlArguments(object):
    """
    This class is responsible for parsing and normalizing the base arguments
    supplied to a GraphQL query node into the corresponding arguments expected
    by a ravel Query object.
    """

    @staticmethod
    def extract_arguments_dict(ast_node) -> Dict:
        return {
            arg.name: arg.value for arg in
            getattr(ast_node, 'arguments', ())
        }

    @classmethod
    def parse(cls, resource_type, ast_node) -> 'GraphqlArguments':
        args = cls.extract_arguments_dict(ast_node)
        return cls(
            where=cls._parse_where(resource_type, args.pop('where', None)),
            order_by=cls._parse_order_by(args.pop('order_by', None)),
            offset=cls._parse_offset(args.pop('offset', None)),
            limit=cls._parse_limit(args.pop('limit', None)),
            custom=args  # `custom` is whatever remains in args
        )

    @classmethod
    def _parse_order_by(cls, order_by_strs: Union[Text, Tuple[Text]]) -> Tuple[OrderBy]:
        if isinstance(order_by_strs, str):
            order_by_strs = [order_by_strs]
        else:
            order_by_strs = order_by_strs or []

        order_bys = []

        for order_by_str in order_by_strs:
            parts = order_by_str.strip().split()

            if len(parts) == 1:
                key = parts
                desc = False
            elif len(parts) == 2:
                key = parts[0]

                ordering = parts[1].lower()
                if ordering == 'desc':
                    desc = True
                elif ordering == 'asc':
                    desc = False
                else:
                    raise ValueError(
                        f'unrecognized "order by" value: {order_by_str}'
                    )

                desc = True if parts[1].lower() == 'desc' else False
                order_by = OrderBy.load({'key': key, 'desc': desc})
                order_bys.append(order_by)

        return tuple(order_bys)

    @classmethod
    def _parse_offset(cls, raw_offset) -> int:
        offset = None
        if raw_offset is not None:
            offset = max(int(raw_offset), 0)
        return offset

    @classmethod
    def _parse_limit(cls, raw_limit) -> int:
        limit = None
        if raw_limit is not None:
            limit = max(int(raw_limit), 1)
        return limit

    @classmethod
    def _parse_where(
        cls,
        resource_type: Type['Resource'],
        predicate_strings: List[Text],
    ) -> List['Predicate']:
        if isinstance(predicate_strings, str):
            predicate_strings = [predicate_strings]
        return Predicate.reduce_and(*[
            resource_type.ravel.predicate_parser.parse(pred_str)
            for pred_str in (predicate_strings or [])
        ])

    def __init__(self, where, order_by, offset, limit, custom: Dict):
        self.where = where
        self.order_by = order_by
        self.offset = offset
        self.limit = limit
        self.custom = custom

    def __getattr__(self, attr):
        return self.custom.get(attr)



from ravel.exceptions import ManifestError


class ManifestFileNotFound(ManifestError):
    pass


class UnrecognizedManifestFileFormat(ManifestError):
    pass


class ManifestValidationError(ManifestError):
    pass


class ManifestInheritanceError(ManifestError):
    pass


class StoreClassNotFound(ManifestError):
    pass


class ResourceClassNotFound(ManifestError):
    pass


class DuplicateResourceClass(ManifestError):
    pass


class DuplicateStoreClass(ManifestError):
    pass


class FilesystemScanTimeout(ManifestError):
    pass


import os
import re
from threading import current_thread
import pkg_resources
import concurrent.futures

from collections import deque
from concurrent.futures import ThreadPoolExecutor
from typing import Text, Dict, Type, Union, Set, List, Callable
from datetime import datetime
from copy import deepcopy

from appyratus.utils.dict_utils import DictUtils, DictObject
from appyratus.utils.string_utils import StringUtils
from appyratus.utils.type_utils import TypeUtils
from appyratus.files import Yaml, Json
from appyratus.env import Environment

from ravel.exceptions import ManifestError
from ravel.util.misc_functions import get_class_name
from ravel.util.loggers import console
from ravel.schema import Schema, fields

from .scanner import ManifestScanner
from .bootstrap import Bootstrap
from .binding import Binding
from .exceptions import *


class Manifest:

    class Schema(Schema):
        base = fields.FilePath()
        package = fields.String()
        bindings = fields.List(Binding.Schema(), default=[])
        bootstraps = fields.List(Bootstrap.Schema(), default=[])
        values = fields.Dict(default={})
        logging = fields.Nested({
            'level': fields.Enum(fields.String(), {
                'DEBUG', 'INFO', 'WARNING', 'CRITICAL', 'ERROR'
            }, default='DEBUG')
        })

    def __init__(self, source: Union['Manifest', Dict, Text, Callable]):
        filepath, data = self._load_manifest_data(source)

        self.env = Environment()
        self.filepath = filepath
        self.data = data
        self.bindings = []
        self.bootstraps = []
        self.default_bootstrap = None
        self.values = None
        self.scanner = None
        self.app = None

    def bootstrap(
        self,
        app: 'Application',
        namespace: Dict = None
    ) -> 'Manifest':

        console.debug(
            message='computed manifest...',
            data=self.data
        )

        self.app = app
        self.bootstraps = self._initialize_bootstraps()
        self.bindings = self._initialize_bindings()
        self.scanner = self._scan_filesystem()

        if namespace:
            self._scan_namespace_dict(namespace)

        self._post_process_values_dict()
        self._resolve_bindings()
        self._resolve_resource_id_fields()
        self._bootstrap_store_classes()
        self._bootstrap_resource_classes()
        self._bind_resources()
        self._inject_classes_to_actions()
        self._bootstrap_app_actions()

        return self

    def register(
        self,
        resource_classes: Union['Resource', List[Type['Resource']]],
        store_class: Type['Store'] = None
    ):
        """
        Dynamically register and bootstrap a resource with its associated
        store class.
        """
        from ravel.resource import Resource

        def resolve_store_class(resource_class, store_class):
            if not store_class:
                if self.default_bootstrap:
                    return self.default_bootstrap.store_class
                else:
                    return resource_class.__store__()
            else:
                return store_class

        def validate_resource_class(resource_class) -> Type['Resource']:
            if not isinstance(resource_class, Resource):
                raise TypeError(
                    f'{get_class_name(resource_class)} must '
                    f'be a Resource subclass'
                )
            if resource_class in self.resource_classes:
                raise DuplicateResourceClass(
                    f'{get_class_name(resource_class)} is already registerd '
                    f'with the application'
                )
            return resource_class

        def create_and_add_binding(resource_class, store_class):
            resource_class_name = get_class_name(resource_class)
            store_class_name = get_class_name(store_class)
            binding = Binding(resource_class_name, store_class_name)
            binding.resource_class = resource_class
            binding.store_class = store_class
            self.bindings.append(binding)
            return binding

        def register_classes(resource_class, store_class):
            # add the resource class to global register if it's new
            resource_class_name = get_class_name(resource_class)
            if resource_class_name not in self.resource_classes:
                self.resource_classes[resource_class] = resource_class

            # add the store class to global register if it's new
            store_class_name = get_class_name(store_class)
            if store_class_name not in self.store_classes:
                self.store_classes[store_class_name] = store_class

        def bootstrap_new_classes(resource_class, store_class):
            resource_class_name = get_class_name(resource_class)
            store_class_name = get_class_name(store_class)
            bootstraps = {
                x.store_class_name: x for x in self.bootstraps
            }
            if self.app.is_bootstrapped:
                bootstrap = self.bootstraps.get(store_class_name)
                params = bootstrap.params if bootstrap else {}
                if not store_class.is_bootstrapped():
                    store_class.bootstrap(self.app, **params)
                if not resource_class.is_bootstraped():
                    resource_class.bootstrap(self.app, **params)

        # make sure we don't already have a store class registered
        if store_class in self.store_classes:
            raise DuplicateStoreClass(
                f'{get_class_name(store_class)} is already registerd '
                f'with the application'
            )

        # register new classes and bootstrap
        for resource_class in resource_classes:
            resource_class = validate_resource_class(resource_class)
            store_class = resolve_store_class(resource_class, store_class)

            register_classes(resource_class, store_class)

            binding = create_and_add_binding(resource_class, store_class)

            # bind new store singleton to resource_class
            store = store_class()
            resource_class.bind(store)
            store.bind(resource_class)

            # bootstrap the resource and store classes if they are
            # not alredy but the host app is.
            bootstrap_new_classes(resource_class, store_class)

    @property
    def package(self) -> Text:
        return self.data.get('package')

    @property
    def logging(self) -> Dict:
        return self.data.get('logging', {
            'level': 'DEBUG'
        }).copy()

    @property
    def resource_classes(self) -> Dict[Text, Type['Resource']]:
        if self.scanner is not None:
            return self.scanner.context.resource_classes
        return DictObject()

    @property
    def store_classes(self) -> Dict[Text, Type['Store']]:
        if self.scanner is not None:
            return self.scanner.context.store_classes
        return DictObject()

    def _resolve_bindings(self):
        def load(binding):
            b = binding

            resource_class = self.resource_classes.get(b.resource_class_name)
            b.resource_class = resource_class
            if resource_class is None:
                raise ResourceClassNotFound(b.resource_class_name)

            store_class = self.store_classes.get(b.store_class_name)
            b.store_class = store_class
            if store_class is None:
                raise StoreClassNotFound(b.store_class_name)

            return binding

        # if the abstract base Resource class is present, dispose with it!
        self.resource_classes.pop('Resource', None)

        for binding in self.bindings:
            binding = load(binding)

    def _bind_resources(self):
        for binding in self.bindings:
            store = binding.store_class()
            store.bind(binding.resource_class, **binding.bind_params)
            binding.resource_class.bind(store, **binding.bind_params)

    def _resolve_resource_id_fields(self):
        """
        Resolve the concrete Field class to use for each "foreign key"
        Id field referenced declared in resource classes. This should happen
        prior to bootstrapping Stores, as without this step, they won't know
        what to make of the unresolved Id fields.
        """
        for resource_class in self.resource_classes.values():
            for id_field in resource_class.ravel.foreign_keys.values():
                id_field.replace_self_in_resource_type(
                    self.app, resource_class
                )

    def _bootstrap_resource_classes(self):
        app = self.app
        for binding in self.bindings:
            binding.resource_class.bootstrap(app)

    def _inject_classes_to_actions(self):
        for action in self.app.actions.values():
            self.app.inject(action.target)

    def _bootstrap_app_actions(self):
        for action in self.app.actions.values():
            action.bootstrap()

        on_parts = []
        pre_parts = []
        post_parts = deque()

        def is_virtual(func):
            return getattr(func, 'is_virtual', None)

        # middleware bootstrapping...
        for idx, mware in enumerate(self.app.middleware):
            mware.bootstrap(app=self.app)

            # everything below is for generating the log message
            # containing the "action execution diagram"
            name = StringUtils.snake(get_class_name(mware))
            if not is_virtual(mware.pre_request):
                pre_parts.append(
                    f"↪ {name}.pre_request(raw_args, raw_kwargs)"
                )

            if not is_virtual(mware.on_request):
                on_parts.append(
                    f"↪ {name}.on_request(args, kwargs)"
                )
            if not is_virtual(mware.post_request):
                if is_virtual(mware.post_bad_request):
                    post_parts.appendleft(
                        f"↪ {name}.post_request(result)"
                    )
                else:
                    post_parts.appendleft(
                        f"↪ {name}.post_[bad_]request(result|error)"
                    )
            elif not is_virtual(mware.post_bad_request):
                post_parts.appendleft(
                    f"↪ {name}.post_bad_request(error)"
                )

        parts = []
        parts.append(
            '➥ app.on_request(action, *raw_args, **raw_kwargs)'
        )
        parts.extend(pre_parts)
        parts.append(
            '➥ args, kwargs = action.marshal(raw_args, raw_kwargs)'
        )
        parts.extend(on_parts)
        parts.append(
            '➥ raw_result = action(*args, **kwargs)'
        )
        parts.extend(post_parts)
        parts.append(
            '➥ return app.on_response(action, raw_result)'
        )

        diagram = '\n'.join(
            f'{"  " * (i+1)}{s}' for i, s in enumerate(parts)
        )
        console.debug(
            message=(
                f"action execution diagram...\n\n {diagram}\n"
            )
        )

    def _bootstrap_store_classes(self):
        app = self.app
        bootstraps = self.bootstraps
        store_classes = self.store_classes

        # if the abstract base Store class is present, dispose with it!
        self.store_classes.pop('Store', None)

        for bootstrap in bootstraps:
            store_class = store_classes.get(bootstrap.store_class_name)
            if store_class is None:
                raise StoreClassNotFound(
                    f'cannot find {bootstrap.store_class_name}'
                )
            else:
                store_class.bootstrap(app, **bootstrap.bootstrap_params)


    def _initialize_bootstraps(self):
        return [
            Bootstrap(x['store'], x.get('params'), x.get('default'))
            for x in self.data.get('bootstraps', [])
        ]

    def _initialize_bindings(self):
        default_bootstrap = None
        store_class_names = set()
        bootstraps = self.bootstraps
        bindings = []

        if bootstraps:
            for x in bootstraps:
                store_class_names.add(x.store_class_name)
                if x.is_default:
                    default_bootstrap = x

        self.default_bootstrap = default_bootstrap

        for x in self.data.get('bindings', []):
            store_class_name = x.get('store')
            if not store_class_name:
                if default_bootstrap:
                    store_class_name = default_bootstrap.store_class_name
                else:
                    store_class_name = 'SimulationStore'
                    console.warning(
                        f'defaulting {x["resource"]} store to '
                        f'{store_class_name}'
                    )

            params = x.get('params') or {}
            binding = Binding(x['resource'], store_class_name, params)
            bindings.append(binding)

            if binding.store_class_name not in store_class_names:
                bootstrap = Bootstrap(binding.store_class_name)
                bootstraps.append(bootstrap)

        return bindings

    @classmethod
    def _load_manifest_data(cls, source):
        schema = cls.Schema(allow_additional=True)
        filepath = None
        data = {}

        # if source is callable, it means that the manifest is being returned
        # by a function, allowing for dynamic manifest properties
        if callable(source):
            source = source()

        # load the manifest data dict differently, depending
        # on its source -- e.g. from a file path, a dict, another manifest
        if isinstance(source, Manifest):
            data = deepcopy(source.data)
            filepath = source.filepath
        elif isinstance(source, str):
            filepath = os.path.abspath(
                os.path.expandvars(os.path.expanduser(source))
            )
            if os.path.isfile(source):
                console.debug(f'reading manifest from {source}')
                data = cls._read_file(source)
            else:
                raise ManifestFileNotFound(
                    f'manifest file {filepath} not found. '
                    f'yaml and json manifest file types are supported.'
                )
        elif isinstance(source, dict):
            data = source
            filepath = None

        # merge data dict into recursively inherited data dict
        base_filepath = data.get('base')
        if base_filepath:
            inherited_data = cls.inherit_base_manifest(base_filepath)
            data = DictUtils.merge(inherited_data, data)

        data = cls._expand_vars(data)

        # validate final computed data dict
        validated_data, errors = schema.process(data)
        if errors:
            raise ManifestValidationError(
                f'manifest validation error/s: {errors}'
            )

        return (filepath, validated_data)

    @classmethod
    def _read_file(cls, path: Text) -> Dict:
        ext = os.path.splitext(path)[-1].lower().lstrip('.')
        if ext in ('yaml', 'yml'):
            return cls._expand_vars(Yaml.read(path) or {})
        elif ext == 'json':
            return cls._expand_vars(Json.read(path) or {})
        else:
            raise UnrecognizedManifestFileFormat(
                f'only json and yaml manifest files are '
                f'supported: {path}'
            )

    @classmethod
    def inherit_base_manifest(
        cls,
        base: Text = None,
        visited: Set = None,
        data: Dict = None,
    ):
        """
        If there is a base manifest file specified by the loaded manifest
        data, then load it here and merge the current manifest into the base
        manifest, recusively back to the root base.
        """
        visited = set() if visited is None else visited
        data = data if data is not None else {}
        base = os.path.abspath(
            os.path.expandvars(
                os.path.expanduser(base)
            )
        )

        if base in visited:
            # avoid infinite loop if there is a bad inheritence cycle
            raise ManifestInheritanceError(
                f'manifest inheritance loop detected: {base}'
            )
        else:
            visited.add(base)

        console.debug(
            f'inheriting manifest from {base}'
        )

        if os.path.isfile(base):
            # merge current data dict into new base dict
            base_data = cls._read_file(base)
            data = DictUtils.merge(base_data.copy(), data)
            # recurse on base's base manifest...
            nested_base = base_data.get('base')
            if nested_base:
                cls.inherit_base_manifest(nested_base, visited, data)

        return data

    @classmethod
    def _expand_vars(cls, data: Dict) -> Dict:
        """
        Recursively expand any environment variable (with the form $FOO or
        ${FOO}) that appears in the data dict, either as a key or value, with
        its corresponding value. Return a new dict.
        """
        expanded = {}
        for k, v in data.items():
            k = k.strip()
            if isinstance(k, str) and k.startswith('$'):
                k = os.path.expandvars(k)
            if isinstance(v, str):
                expanded[k] = os.path.expandvars(v.strip())
            elif isinstance(v, dict):
                expanded[k] = cls._expand_vars(v)
            elif isinstance(v, list):
                expanded[k] = [cls._expand_vars(lv) for lv in v]
            else:
                expanded[k] = v
        return expanded

    def _scan_filesystem(self) -> ManifestScanner:
        """
        Use venusian simply to scan the action packages/modules, causing the
        action callables to register themselves with the Application instance.
        """
        t1 = datetime.now()
        package = self.package
        executor = ThreadPoolExecutor(
            max_workers=6,
            thread_name_prefix=(
                f'{StringUtils.camel(self.package or "")}'
                f'ManifestWorker'.strip()
            )
        )
        scanner = ManifestScanner(self)
        futures = []

        def scan(package, verbose=False):
            console.debug(f'manifest scanning {package}')
            try:
                scanner.scan(package)
            except:
                console.exception('scan failed')

        def async_scan(package, verbose=False):
            future = executor.submit(scan, package, verbose)
            futures.append(future)

        async_scan('ravel.resource')
        async_scan('ravel.store')

        installed_pkg_names = {
            pkg.key for pkg in pkg_resources.working_set
        }

        # scan extension directories if the package installation requirements
        # are met, like sqlalchemy and redis.
        pgk_name_2_ravel_scan_path = {
            'sqlalchemy': 'ravel.ext.sqlalchemy',
            'redis': 'ravel.ext.redis',
            'pygame': 'ravel.ext.gaming.pygame',
            'celery': 'ravel.ext.celery',
            'numpy': 'ravel.ext.np'
        }
        for pkg_name, scan_path in pgk_name_2_ravel_scan_path.items():
            if pkg_name in installed_pkg_names:
                async_scan(scan_path)

        # scan the app project package
        if package:
            async_scan(package, verbose=True)

        completed_scans, incomplete_scans = (
            concurrent.futures.wait(futures)
        )
        if incomplete_scans:
            raise FilesystemScanTimeout(
                message='filesystem scan timed out',
            )

        t2 = datetime.now()
        secs = (t2 - t1).total_seconds()
        console.debug(f'scanned filesystem in {secs:.2f}s')

        return scanner

    def _scan_namespace(self, namespace: Dict):
        """
        Non-recursively detect Store and Resource class objects contained in
        the given namespace dict, making them available to the bootstrapping
        app.
        """
        from ravel.store import Store
        from ravel.resource import Resource

        for k, v in (namespace or {}).items():
            if TypeUtils.is_proper_subclass(v, Resource):
                if not v.ravel.is_abstract:
                    self.resource_classes[k] = v
                    console.debug(
                        f'detected Resource class in '
                        f'namespace dict: {get_class_name(v)}'
                    )
            elif TypeUtils.is_proper_subclass(v, Store):
                self.store_classes[k] = v
                console.debug(
                    f'detected Store class in namespace '
                    f'dict: {get_class_name(v)}'
                )

    def _post_process_values_dict(self):
        """
        Recursively convert data['values'] into a tree of DictObjects
        """
        def objectify(data):
            copy = data.copy()
            for k, v in data.items():
                if isinstance(v, dict):
                    copy[k] = objectify(v)
            return DictObject(copy)

        self.values = objectify(self.data['values'])


from typing import Tuple, Text

from ravel.util.loggers import console
from ravel.util.misc_functions import get_class_name
from ravel.util import is_resource, is_batch, is_sequence
from ravel.query.order_by import OrderBy
from ravel.query.request import Request

DNE = {}


class ResolverProperty(property):
    """
    # ResolverProperty
    The resolvers declared on resource classes are dynamically converted into resolver
    properties at runtime. Resolving, reading, and writing resource instance state is
    managed through ResolverDecorator.

    ## Lazy Loading
    If necessary, ResolverProperty lazy loads its value on "get" by executing
    its resolver's resolve method.

    ## Resolver CRUD Callbacks
    The getter, setter, and deleter of ResolverProperty execute its resolver's
    on_get, on_set, and on_delete callbacks.
    """

    def __init__(
        self,
        resolver: 'Resolver',
        decorator: 'ResolverDecorator' = None
    ):
        """
        Args:
        - `resolver`: The resolver managed by this property.
        - `decorator`: The ResolverDecorator used to build the resolver, if any.
        """
        super().__init__(fget=self.fget, fset=self.fset, fdel=self.fdel)
        self.decorator = decorator
        self.resolver = resolver

    @property
    def app(self) -> 'Application':
        """
        The host Application for this property's resolver.
        """
        return self.resolver.owner.ravel.app

    def select(self, *items: Tuple):
        """
        Build and return a query request. This is used in query-building syntax.
        It's used like this:

        ```python
        account_request = User.account.select(Account._id)
        query = User.select(User._id, request)

        # or more succinctly:
        query = User.select(
            User._id,
            User.account.select(Account._id),
        )
        ```
        """
        request = Request(self.resolver)
        request.select(*items)
        self.resolver.on_select(request)
        return request

    def fget(self, resource: 'Resource'):
        """
        Returns resource state data associated with the resolver. For example,
        if the name of the resolver were "email", corresponding to an email
        field, we would return `resource.state["email"]`. However, if "email" is
        missing from the state dict, then we lazy load it through the resolver,
        memoizing it in the state dict. This method also runs the resolver's
        on_get callback.
        """
        resolver = self.resolver

        # if value not loaded, lazily resolve it
        if resolver.name not in resource.internal.state:
            console.debug(f'lazy loading {resource}.{resolver.name}')
            request = Request(resolver)
            value = resolver.resolve(resource, request)

            # if resolver is for a field, we know that the resolved
            # field has been loaded ON the returned resource (value)
            if resolver.name in resource.ravel.resolvers.fields:
                value = resource.internal.state.get(resolver.name, DNE)
                if value is DNE:
                    console.debug(
                        message=(
                            f'no value returned for '
                            f'{resource}.{resolver.name}'
                        )
                    )
                    return

            if (value is not None) or resolver.nullable:
                resource.internal.state[resolver.name] = value
            elif (value is None) and (not resolver.nullable):
                if resource.internal.state.get(resolver.name) is None:
                    resource.internal.state.pop(resolver.name, None)
                console.warning(
                    message=(
                        f'resolver returned bad value'
                    ),
                    data={
                        'resource': resource._id,
                        'class': resource.class_name,
                        'resolver': self.resolver.name,
                        'reason': 'resolver not nullable',
                    }
                )

        value = resource.internal.state.get(resolver.name)
        resolver.on_get(resource, value)
        return value

    def fset(self, resource: 'Resource', new_value):
        """
        Set resource state data, calling the resolver's on_set callbak.
        """
        resolver = self.resolver
        target_class = self.resolver.target
        target_schema = target_class.ravel.schema
        old_value = resource.internal.state.get(resolver.name)

        # cast plain lists and dicts to target Resource or Batch objects
        if self.resolver.many:
            if not is_batch(new_value):
                assert is_sequence(new_value)
                new_value = self.resolver.target.Batch(
                    target_class(x) if isinstance(x, dict) else x
                    for x in new_value
                )
        elif isinstance(new_value, dict) \
                and resolver.name not in target_schema.fields:
            new_value = self.resolver.target(new_value)

        # write new value to instance state
        resource.internal.state[resolver.name] = new_value

        # trigger on_set callback
        resolver.on_set(resource, old_value, new_value)

    def fdel(self, resource: 'Resource'):
        """
        Remove state data from the resource's state dict, calling the resolver's
        on_delete callback.
        """
        resolver = self.resolver
        if resolver.name in resource.internal.state:
            old_value = resource.internal.state.pop(resolver.name)
            resolver.on_delete(resource, old_value)



from __future__ import absolute_import

import six

from falcon import errors
from falcon.media import BaseHandler
from falcon.util import json

from ravel.util.json_encoder import JsonEncoder


class JsonHandler(BaseHandler):
    """
    # Json Handler
    """

    def __init__(self, encoder: 'JsonEncoder' = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoder = encoder or JsonEncoder()

    def deserialize(self, stream, content_type, content_length):
        try:
            raw = stream.read()
            return self.encoder.decode(raw.decode('utf-8'))
        except ValueError as err:
            raise errors.HTTPBadRequest(
                'Invalid JSON',
                'Could not parse JSON body - {0}'.format(err)
            )

    def serialize(self, media, content_type):
        result = self.encoder.encode(media)
        if six.PY3 or not isinstance(result, bytes):
            result = result.encode('utf-8')
        return result



DEFAULT_HOST = '127.0.0.1'
DEFAULT_PORT = '50051'
DEFAULT_GRACE = 5.0
DEFAULT_IS_SECURE = False
REGISTRY_PROTO_FILE = 'app.proto'
PB2_MODULE_PATH_FSTR = '{}.grpc.app_pb2'
PB2_GRPC_MODULE_PATH_FSTR = '{}.grpc.app_pb2_grpc'
PROTOC_COMMAND_FSTR = '''
python3 -m grpc_tools.protoc
    -I {include_dir}
    --python_out={build_dir}
    --grpc_python_out={build_dir}
    {proto_file}
'''
PROTOC_WEB_GENERATE_MESSAGE_CLASSES_COMMAND = '''
protoc 
    -I {include_dir}
    --js_out=import_style=commonjs:{build_dir}
    {proto_file}
'''
PROTOC_WEB_GENERATE_CLIENT_COMMAND = '''
protoc
    -I {include_dir}
    --grpc-web_out=import_style=commonjs,mode=grpcwebtext:{build_dir}
    {proto_file}
'''


import bisect
import time

import numpy as np
import BTrees.OOBTree

from copy import deepcopy
from collections import defaultdict, Counter
from threading import RLock
from functools import reduce
from typing import Text, Dict, List, Set, Tuple, Type

from BTrees.OOBTree import BTree

from appyratus.utils.dict_utils import DictUtils

from ravel.schema import Schema, Field, fields
from ravel.constants import ID, REV
from ravel.util import union
from ravel.query.order_by import OrderBy
from ravel.query.predicate import (
    Predicate,
    ConditionalPredicate,
    BooleanPredicate,
    OP_CODE,
)

from .base import Store


class SimulationStore(Store):
    """
    An in-memory Store that stores data in Python dicts with BTrees indexes.
    """

    def __init__(self):
        super().__init__()
        self.reset()

    def on_bind(self, resource_type: Type['Resource'], **kwargs):
        for k, field in resource_type.ravel.schema.fields.items():
            if field.scalar and (type(field) is not Field):
                self.indexes[k] = BTree()

    def reset(self):
        """
        Reset all internal data structures.
        """
        self.lock = RLock()
        self.indexes = {}
        self.records = {}

    def exists(self, _id) -> bool:
        """
        Does the _id exist in the store?
        """
        with self.lock:
            return _id in self.records

    def exists_many(self, _ids: List) -> Dict[object, bool]:
        with self.lock:
            return {
                _id: (_id in self.records)
                for _id in _ids
            }

    def count(self) -> int:
        """
        Return the total number of objects in the store.
        """
        with self.lock:
            return len(self.records)

    def fetch(self, _id, fields: Set[Text] = None) -> Dict:
        """
        Return a single record.
        """
        return self.fetch_many([_id], fields=fields).get(_id)

    def fetch_many(self, _ids: List, fields=None) -> Dict:
        """
        Return multiple records in a _id-keyed dict.
        """
        if not fields:
            fields = set(self.resource_type.ravel.schema.fields)
        elif not isinstance(fields, set):
            fields = set(fields)

        with self.lock:
            records = {}

            for _id in _ids:
                # return a *copy* so as not to mutate dict in store
                record = deepcopy(self.records.get(_id))
                records[_id] = record

                # remove keys not specified in "fields" argument
                if fields and (record is not None):
                    record_keys = set(record.keys())
                    for k in record_keys - fields:
                        del record[k]

            return records

    def fetch_all(self, fields=None) -> Dict:
        """
        Return all records in a _id-keyed dict.
        """
        with self.lock:
            return deepcopy(self.records)

    def create(self, record: Dict = None) -> Dict:
        """
        Insert one record into the store, indexing its indexable fields.
        """
        schema = self.resource_type.ravel.schema

        with self.lock:
            record[ID] = self.create_id(record)
            record[REV] = self.increment_rev()

            _id = record[ID]

            # insert the record and update indexes for its fields
            self.records[_id] = record
            self._index_upsert(_id, record)

        return deepcopy(record)

    def create_many(self, records: List[Dict] = None) -> List[Dict]:
        """
        Insert multiple records into the store, indexing their indexable
        fields.
        """
        results = []
        with self.lock:
            for record in records:
                record[ID] = self.create_id(record)
                result = self.create(record)
                results.append(result)
        return results

    def update(self, _id=None, data: Dict = None) -> Dict:
        """
        Submit changes to an object in the store.
        """
        with self.lock:
            old_record = self.records.get(_id, {})
            old_rev = old_record.get(REV)

            # remove the old copy of the object from the store, and then replace
            # it. we use `internal` in the call to delete to indicate that we
            # are only deleting for the purpose of re-inserting during this
            # update. this prevents the _rev index from being cleared.
            if old_record:
                self.delete(
                    old_record[ID],
                    index_names=set(data.keys()),
                    internal=True,
                )

            merged_record = DictUtils.merge(old_record, data)
            merged_record[REV] = self.increment_rev(old_rev)

            # "re-insert" the new copy of the record
            record = self.create(merged_record)
            return record

    def update_many(self, _ids: Set, data: Dict = None) -> Dict:
        """
        Submit changes to multiple objects in the store.
        """
        with self.lock:
            return {
                _id: self.update(_id=_id, data=x)
                for _id, x in zip(_ids, data)
            }

    def delete(self, _id, index_names: Set[Text] = None, internal=False):
        """
        Delete one record along with its field indexes.
        """
        with self.lock:
            record = self.records.get(_id)
            if record:
                index_names = set(index_names or record.keys())
                if internal:
                    index_names.discard(REV)
                self._index_remove(_id, record, index_names)
                del self.records[_id]

    def delete_many(self, _ids: List, internal=False):
        """
        Delete multiple records along with their field indexes.
        """
        with self.lock:
            for _id in _ids:
                self.delete(_id, internal=internal)

    def delete_all(self):
        """
        Delete all records along with their field indexes.
        """
        self.delete_many(self.records.keys())

    def query(
        self,
        predicate: Predicate,
        fields: Set[Text] = None,
        order_by: Tuple = None,
        limit: int = None,
        offset: int = None,
        **kwargs
    ) -> List:
        """
        """
        with self.lock:
            # compute set of ID's of records whose fields satisfy
            # the given "where" predicate of the query
            computed_ids = self._eval_predicate(predicate)
            records = list(self.fetch_many(computed_ids, fields).values())

            # order the records
            if order_by:
                records = OrderBy.sort(records, order_by)

            # paginate after ordering
            if offset is not None:
                if limit is not None:
                    records = records[offset:offset+limit]
                else:
                    records = records[offset:]
            elif limit is not None:
                records = records[:limit]

            return records

    def _index_upsert(self, _id, record):
        for k, v in record.items():
            index = self.indexes.get(k)
            if index is not None:
                if v not in index:
                    index[v] = set()
                index[v].add(_id)

    def _index_remove(self, _id, record, index_names=None):
        index_names = set(index_names or record.keys())
        for k in index_names:
            index = self.indexes.get(k)
            if index is not None and k in record:
                if k in record:
                    v = record[k]
                    if v in index:
                        index[v].remove(_id)

    def _eval_predicate(self, predicate):
        if predicate is None:
            return self.records.keys()

        op = predicate.op
        empty = set()
        computed_ids = set()

        if isinstance(predicate, ConditionalPredicate):
            k = predicate.field.source
            v = predicate.value
            index = self.indexes[k]

            if op == OP_CODE.EQ:
                computed_ids = index.get(v, empty)
            elif op == OP_CODE.NEQ:
                computed_ids = union([
                    id_set for v_idx, id_set in index.items()
                    if v_idx != v
                ])
            elif op == OP_CODE.INCLUDING:
                # containment - we compute the union of all sets of ids whose
                # corresponding records have the given values in the index
                v = v if isinstance(v, set) else set(v)
                computed_ids = union([index.get(k_idx, empty) for k_idx in v])

            elif op == OP_CODE.EXCLUDING:
                # the inverse of containment...
                v = v if isinstance(v, set) else set(v)
                computed_ids = union([
                    id_set for v_idx, id_set in index.items()
                    if v_idx not in v
                ])
            else:
                # handle inequalities, computing limit and offset to form an
                # interval with which we index the actual BTree

                keys = np.array(index.keys(), dtype=object)

                if op == OP_CODE.GEQ:
                    offset = bisect.bisect_left(keys, v)
                    interval = slice(offset, None, 1)

                elif op == OP_CODE.GT:
                    offset = bisect.bisect(keys, v)
                    interval = slice(offset, None, 1)

                elif op == OP_CODE.LT:
                    offset = bisect.bisect_left(keys, v)
                    interval = slice(0, offset, 1)

                elif op == OP_CODE.LEQ:
                    offset = bisect.bisect(keys, v)
                    interval = slice(0, offset, 1)
                else:
                    # XXX: raise StoreError
                    raise Exception('unrecognized op')

                computed_ids = union([
                    index[k] for k in keys[interval] if k is not None
                ])

        elif isinstance(predicate, BooleanPredicate):
        # recursively compute and union child predicates,
        # left-hand side (lhs) and right-hand side (rhs)
            lhs = predicate.lhs
            rhs = predicate.rhs

            if op == OP_CODE.AND:
                lhs_result = self._eval_predicate(lhs)
                if lhs_result:
                    rhs_result = self._eval_predicate(rhs)
                    computed_ids = set.intersection(lhs_result, rhs_result)

            elif op == OP_CODE.OR:
                lhs_result = self._eval_predicate(lhs)
                rhs_result = self._eval_predicate(rhs)
                computed_ids = set.union(lhs_result, rhs_result)

            else:
                # XXX: raise StoreError
                raise Exception(f'unrecognized predicate operator: {op}')

        return computed_ids



from appyratus.json import JsonEncoder as BaseJsonEncoder

from ravel.util import is_resource, is_batch


class JsonEncoder(BaseJsonEncoder):
    def default(self, target):
        if is_resource(target) or is_batch(target):
            return target.dump()
        else:
            return super().default(target)



from ravel.app.middleware import Middleware


class ManageCeleryClient(Middleware):
    def __init__(self, service: 'CeleryService', mocked=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._celery_service = service
        self._mocked = mocked

    def on_bootstrap(self):
        """
        Initialize a client for the given CeleryService
        """
        self.app.local.celery = self._celery_service.client.bootstrap(
            manifest=self.app.manifest,
            mocked=self._mocked,
        )



import traceback

from typing import Dict, List, Type, Text, Tuple
from collections import deque

from appyratus.utils.time_utils import TimeUtils

from ravel.constants import ID

READ_METHODS = frozenset({
    'fetch', 'fetch_all', 'fetch_many', 'count', 'exists', 'query'
})

WRITE_METHODS = frozenset({
    'create', 'create_many', 'update', 'update_many', 'delete',
    'delete_many', 'delete_all'
})


class StoreEvent(object):
    def __init__(
        self,
        method: Text,
        args: Tuple = None,
        kwargs: Dict = None,
        result = None,
        exc: Exception = None,
    ):
        self.method = method
        self.args = list(args) if args else []
        self.kwargs = kwargs or {}
        self.result = result
        self.timestamp = TimeUtils.utc_now()
        self.exc = exc

        # fix up create and create_many args and kwargs.
        # see the doc string for _backfill_id_fields.
        if (exc is None) and (method in {'create', 'create_many'}):
            self._backfill_id_fields(result)

    def __repr__(self):
        return (
            f'StoreEvent(method={self.method}, timestamp={self.timestamp})'
        )

    def _backfill_id_fields(self, result):
        """
        If we ever replay this history, we don't want to inadvertantly trigger
        the generation of new resource ID's in the store we're replaying it in.
        To prevent this, we inject the generated ID's from the return values of
        these methods into the args and kwargs that were originally passed in.
        """
        if self.method == 'create':
            self.args[0][ID] = result.get(ID)
        elif self.method == 'create_many':
            uncreated_records = args[0]
            created_records = result
            for uncreated, created in zip(uncreated_records, created_records):
                created_id = created_record.get(ID)
                uncreated_record[ID] = created_id

    def dump(self, with_result=False) -> Dict:
        data = {
            'timestamp': self.timestamp,
            'method': self.method,
            'args': self.args,
            'kwargs': self.kwargs,
            'exc': self.exc,
        }
        if with_result:
            data['result'] = self.result
        return data


class StoreHistory(object):
    def __init__(self, store: 'Store', reads=True, writes=True):
        self._store = store
        self._events = deque()
        self._is_recording = False
        self._is_recording_reads = reads
        self._is_recording_writes = writes

    def __len__(self):
        return len(self._events)

    def __iter__(self):
        return iter(self._events)

    def __getitem__(self, idx):
        return self._events[idx]

    def is_recording_method(self, name):
        return (
            self._is_recording and (
                (name in READ_METHODS and self._is_recording_reads) or
                (name in WRITE_METHODS and self._is_recording_writes)
            )
        )

    def start(self):
        self._is_recording = True

    def stop(self):
        self._is_recording = False

    def clear(self):
        self._events = []

    def append(self, event):
        self._events.append(event)

    @property
    def is_recording(self):
        return self._is_recording

    @property
    def events(self):
        return self._events



from uuid import UUID
from datetime import datetime
from threading import local
from types import GeneratorType

from redis import StrictRedis
from appyratus.utils.time_utils import TimeUtils


class RedisClient(StrictRedis):
    pass


class RedisObject(object):
    def __init__(self, redis, name):
        self._redis = redis
        self._name = name

    @property
    def redis(self):
        return self._redis

    @property
    def name(self):
        return self._name


class HashSet(RedisObject):
    def __contains__(self, key):
        return self.redis.hexists(self.name, key)

    def __iter__(self):
        return self.keys()

    def __getitem__(self, key):
        return self.redis.hget(self.name, key)

    def __setitem__(self, key, value):
        self.redis.hset(self.name, key, value)

    def __delitem__(self, key):
        self.redis.hdel(self.name, key)

    def __len__(self):
        return self.redis.hlen(self.name)

    def increment(self, key, delta=1, pipe=None):
        redis = pipe if pipe is not None else self.redis
        return redis.hincrby(self.name, key, delta)

    def delete(self, key, pipe=None) -> bool:
        redis = pipe if pipe is not None else self.redis
        return bool(redis.hdel(self.name, key))

    def delete_many(self, keys, pipe=None) -> int:
        if isinstance(keys, GeneratorType):
            keys = tuple(keys)
        if keys:
            redis = pipe if pipe is not None else self.redis
            return redis.hdel(self.name, *keys)
        return

    def keys(self, pipe=None):
        redis = pipe if pipe is not None else self.redis
        return (k.decode() for k in redis.hkeys(self.name))

    def values(self, pipe=None):
        redis = pipe if pipe is not None else self.redis
        return (v.decode() for v in redis.hvals(self.name))

    def items(self, pipe=None):
        redis = pipe if pipe is not None else self.redis
        return zip(self.keys(pipe=pipe), self.values(pipe=pipe))

    def update(self, mapping, pipe=None):
        redis = pipe if pipe is not None else self.redis
        if mapping:
            redis.hmset(self.name, mapping)

    def get(self, key, default=None, pipe=None):
        redis = pipe if pipe is not None else self.redis
        return redis.hget(self.name, key) or default

    def get_many(self, keys, pipe=None):
        if isinstance(keys, GeneratorType):
            keys = tuple(keys)
        if keys:
            redis = pipe if pipe is not None else self.redis
            return redis.hmget(self.name, *keys)
        else:
            return []

    def get_all(self):
        return self.redis.hgetall(self.name)


class RangeIndex(RedisObject):

    DELIM = '\0\0'
    DELIM_BYTES = DELIM.encode()

    def upsert(self, _id, value, pipe=None):
        raise NotImplementedError('override in subclass')

    def delete(self, _id, pipe=None):
        raise NotImplementedError('override in subclass')

    def delete_many(self, _ids, pipe=None):
        raise NotImplementedError('override in subclass')

    def search(
        self,
        lower=None,
        upper=None,
        include_lower=True,
        include_upper=False,
        offset=None,
        limit=None,
        pipe=None,
    ):
        raise NotImplementedError('override in subclass')


class NumericIndex(RangeIndex):
    custom_serializers = {
        UUID: lambda x: int(x.hex, 16),
        datetime: lambda x: TimeUtils.to_timestamp(x),
        bool: lambda x: int(x),
    }

    def upsert(self, _id, value, pipe=None):
        redis = pipe if pipe is not None else self.redis
        ser = self.custom_serializers.get(value.__class__)
        value = ser(value) if ser else value

        redis.zrem(self.name, _id)
        redis.zadd(self.name, {_id: value})

    def delete(self, _id, pipe=None):
        redis = pipe if pipe is not None else self.redis
        redis.zrem(self.name, _id)

    def delete_many(self, _ids, pipe=None):
        if isinstance(_ids, GeneratorType):
            _ids = tuple(_ids)
        if _ids:
            redis = pipe if pipe is not None else self.redis
            redis.zrem(self.name, *_ids)

    def search(
        self,
        lower=None,
        upper=None,
        include_lower=True,
        include_upper=False,
        offset=None,
        limit=None,
        pipe=None
    ):
        if lower is not None:
            lower = '({}'.format(lower) if not include_lower else lower
        else:
            lower = '-inf'

        if upper is not None:
            upper = '({}'.format(upper) if not include_upper else upper
        else:
            upper = '+inf'

        if limit is not None and offset is None:
            offset = 0

        redis = pipe if pipe is not None else self.redis

        return [
            v.split(self.DELIM_BYTES)[-1]
            for v in redis.zrangebyscore(
                self.name, lower, upper, start=offset, num=limit
            )
        ]


class StringIndex(RangeIndex):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lutab = HashSet(self.redis, ':'.join([self.name, 'lutab']))

    def upsert(self, _id, value, pipe=None):
        redis = pipe if pipe is not None else self.redis
        old_key = self.lutab.get(_id)
        new_key = '{}{}{}'.format(value, self.DELIM, _id)

        if old_key is not None:
            redis.zrem(self.name, old_key)

        # TODO: Do this in a pipeline
        redis.zadd(self.name, {new_key: 0.0})
        self.lutab[_id] = new_key

    def delete(self, _id, pipe=None):
        old_key = self.lutab.get(_id)
        if old_key is not None:
            redis = pipe if pipe is not None else self.redis
            self.lutab.delete(old_key)
            redis.zrem(self.name, old_key)

    def delete_many(self, _ids, pipe=None):
        old_keys = self.lutab.get_many(_ids)
        if old_keys:
            redis = pipe if pipe is not None else self.redis
            self.lutab.delete_many(old_keys)
            redis.zrem(self.name, *old_keys)

    def search(
        self,
        lower=None,
        upper=None,
        include_lower=False,
        include_upper=True,
        offset=None,
        limit=None,
        pipe=None,
    ):
        if lower is None:
            lower = '-'
        else:
            lower = '[' + lower + self.DELIM
            if not include_lower:
                lower += '\xff'

        if upper is None:
            upper = '+'
        else:
            upper = '[' + upper + self.DELIM
            if include_upper:
                upper += '\xff'

        if limit is not None and offset is None:
            offset = 0

        redis = pipe if pipe is not None else self.redis
        values = redis.zrangebylex(
            self.name, lower, upper, start=offset, num=limit
        )

        return [
            v.split(self.DELIM_BYTES)[-1] for v in values
        ]



from __future__ import absolute_import

import traceback

import falcon

from appyratus.env import Environment

from ravel.app.middleware import Middleware
from ravel.apps.web import AbstractWsgiService
from ravel.util.json_encoder import JsonEncoder
from ravel.util.loggers import console

from .media import JsonHandler
from .resource import FalconResource


class FalconService(AbstractWsgiService):

    env = Environment()

    def __init__(self, router=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._json_encoder = JsonEncoder()
        self._route_2_resource = {}
        self._router = router

    @property
    def falcon_middleware(self):
        return []

    @property
    def falcon_resources(self):
        return self._route_2_resource

    @property
    def request_type(self):
        return Request

    @property
    def response_type(self):
        return Response

    @staticmethod
    def handle_error(exc, request, response, params):
        response.status = falcon.HTTP_500
        traceback.print_exc()

    def on_start(self, *args, **kwargs):
        console.info(
            message='starting Falcon web service',
            data={'endpoints': sorted(self.actions.keys())}
        )
        return self.entrypoint(*args, **kwargs)

    def entrypoint(
        self, environ=None, start_response=None, *args, **kwargs
    ):
        middleware = self.falcon_middleware
        for m in middleware:
            if isinstance(m, Middleware):
                m.bind(self)

        falcon_app = falcon.App(
            middleware=middleware,
            request_type=self.request_type,
            response_type=self.response_type,
            router=self._router
        )
        falcon_app.req_options = Request.Options()
        falcon_app.resp_options = Response.Options()
        falcon_app.add_error_handler(Exception, self.handle_error)

        for route, resource in self._route_2_resource.items():
            falcon_app.add_route(route, resource)

        return falcon_app

    def on_decorate(self, endpoint):
        super().on_decorate(endpoint)
        for route in endpoint.routes:
            resource = self._route_2_resource.get(route)
            if resource is None:
                resource = FalconResource(route)
                self._route_2_resource[route] = resource
                resource.add_endpoint(endpoint)
            else:
                resource.add_endpoint(endpoint)

    def on_request(self, endpoint, request, response, *args, **kwargs):
        if request.content_length:
            app_kwargs = dict(request.media or {}, **kwargs)
        else:
            app_kwargs = dict(kwargs)

        app_kwargs.update(request.params)

        # append URL path variables
        route = request.path.strip('/').split('/')
        route_template = request.uri_template.strip('/').split('/')
        for k, v in zip(route_template, route):
            if k and (k[0] == '{' and k[-1] == '}'):
                app_kwargs[k[1:-1]] = v

        return (tuple(), app_kwargs)

    def on_response(
        self,
        action,
        result,
        raw_args=None,
        raw_kwargs=None,
        *args,
        **kwargs
    ):
        request, response = raw_args
        response.media = result
        return result


class Request(falcon.Request):
    """
    Specialized HTTP Request object (NOT a Ravel Request type)
    """

    class Options(falcon.RequestOptions):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.media_handlers['application/json'] = JsonHandler()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = None
        self.json = {}


class Response(falcon.Response):
    class Options(falcon.ResponseOptions):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.media_handlers['application/json'] = JsonHandler()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def ok(self):
        status_code = int(self.status[:3])
        return (200 <= status_code < 300)



from typing import Dict, Text


class TemplateRenderer:
    def render(self, template: Text, context: Dict = None) -> Text:
        """
        Render a template to string, interpolating with context.

        Args:
            - template: filename of the template, like example.html
            - context: the template context data dict
        """
        raise NotImplementedError()


from typing import Dict, Set, Text, List, Type, Tuple

import graphql.ast
import graphql.parser

from appyratus.utils.dict_utils import DictObject

from ravel.util.loggers import console
from ravel.resource import Resource
from ravel.query.query import Query
from ravel.query.request import Request

from .graphql_arguments import GraphqlArguments
from .graphql_query import GraphqlQuery


class GraphqlInterpreter(object):
    """
    The purpose of the GraphQlInterpreter is to take a a raw GraphQL query
    string and generate a corresponding ravel Query object.
    """

    def __init__(self, root_resource_type: Type['Resource']):
        if not issubclass(root_resource_type, GraphqlQuery):
            raise ValueError(
                'root resource type must be a GraphqlQuery subclass'
            )
        self._root_resource_type = root_resource_type
        self._ast_parser = graphql.parser.GraphQLParser()

    def interpret(self, graphql_query_string: Text, context=None) -> Query:
        context = context or DictObject()
        root_node = self._parse_graphql_query_ast(graphql_query_string)
        root_node.source = graphql_query_string
        return self._build_query(root_node, self._root_resource_type, context)

    def _parse_graphql_query_ast(self, graphql_query_string: Text):
        graphql_doc = self._ast_parser.parse(graphql_query_string)
        graphql_query = graphql_doc.definitions[0]
        return graphql_query

    def _build_query(
        self,
        node: graphql.ast.Field,
        target_resource_type: Resource,
        context: Dict,
    ) -> Query:
        """
        Recursively build a ravel Query object from the given low-level GraphQL
        AST node returned from the core GraphQL language parser.
        """
        result = self._root_resource_type()
        query = Query(target=self._root_resource_type, sources=[result])

        for child_node in node.selections:
            resolver = result.ravel.resolvers.get(child_node.name)
            if resolver is None:
                raise Exception(f'unrecognized resolver: {child_node.name}')
            else:
                request = self._build_request(child_node, query, resolver)
                query.select(request)

        return query.limit(1)

    def _build_request(self, node, query, resolver):
        schema = resolver.owner.ravel.schema
        resolvers = resolver.target.ravel.resolvers

        name = node.name
        request = Request(resolver, query=query)
        args = GraphqlArguments.parse(resolver.target, node)

        if args.where:
            request.where(args.where)
        if args.order_by:
            request.order_by(args.order_by)
        if args.offset:
            request.offset(args.offset)
        if args.limit:
            request.limit(args.limit)
        if args.custom:
            request.parameters.update(args.custom)

        for child_node in node.selections:
            name = child_node.name
            if name not in resolvers:
                raise Exception(f'unrecognized resolver: {name}')

            if name in schema.fields:
                request.select(name)
            else:
                child_resolver = resolvers[name]
                child_request = self._build_request(
                    child_node, query, child_resolver
                )

                request.select(child_request)

        return request



from typing import Dict, Text, Callable

from appyratus.enum import Enum

from ravel.util.misc_functions import get_class_name
from ravel.exceptions import RavelError

from .argument import ArgumentSpecification

OP_CODE = Enum(AND='AND', OR='OR', NOT='NOT')


class GuardFailure(RavelError):
    def __init__(self, guard, message=None, *args, **kwargs):
        super().__init__(message or 'guard failed', *args, **kwargs)
        self.guard = guard
        self.data['guard'] = {}
        self.data['guard']['class'] = get_class_name(guard)
        self.data['guard']['failed'] = guard.description
        if guard is not guard.root:
            self.data['guard']['description'] = guard.root.description


class Guard(object):
    """
    Subclasses of Guard must implement guard, which determines
    whether a given request is authorized, by inspecting arguments passed into
    a corresponding Action at runtime.

    Positional and keyword argument names declared in the guard
    method are plucked from the incoming arguments dynamically (following the
    required `request` dict argument).

    The request dict is shared by all Guards composed in a
    CompositeGuard boolean expression.
    """

    def __init__(
        self,
        execute: Callable = None,
        parent: 'Guard' = None,
        **kwargs
    ):
        self.parent = parent
        self.callback = execute
        self.kwargs = kwargs
        if self.callback is not None:
            self.spec = ArgumentSpecification(self.callback)
        else:
            self.spec = ArgumentSpecification(self)

    def __repr__(self):
        return f'{get_class_name(self)}(description=\'{self.description}\')'

    def __call__(self, request: Dict, arguments: Dict) -> bool:
        args, kwargs = self.spec.extract(arguments)
        return self.execute(request, *args, **kwargs)

    def __and__(self, other) -> 'CompositeGuard':
        composite_guard = CompositeGuard(OP_CODE.AND, self, other)
        self.parent = composite_guard
        other.parent = composite_guard
        return composite_guard

    def __or__(self, other) -> 'CompositeGuard':
        composite_guard = CompositeGuard(OP_CODE.OR, self, other)
        self.parent = composite_guard
        other.parent = composite_guard
        return composite_guard

    def __invert__(self) -> 'CompositeGuard':
        composite_guard = CompositeGuard(OP_CODE.NOT, self, None)
        self.parent = composite_guard
        other.parent = composite_guard
        return composite_guard

    def fail(self, message=None) -> GuardFailure:
        return GuardFailure(self, message=message, logged_traceback_depth=1)

    @property
    def description(self) -> Text:
        return get_class_name(self)

    @property
    def root(self) -> 'Guard':
        child = self
        while True:
            if child.parent is None:
                return child
            else:
                child = child.parent

    def execute(self, request: 'Request', *args, **kwargs) -> bool:
        """
        Determine whether Action request is authorized by performing any
        necessary authorization check here. Each subclass must explicitly
        declare which arguments are required. For example,

        ```python
        class UserOwnsPost(Guard):
            def execute(request, user, post):
                return user.owns(post)
        ```

        This implementation expects to be used with a Action with "user"
        and "post" arguments, for instance:

        ```python3
        @repl(auth=UserOwnsPost())
        def delete_post(user, post):
            post.delete()
        ```

        It is possible to combine Guards in boolean expressions, using
        '&' (AND), '|' (OR) and '~' (NOT), like

        ```python3
        @repl(auth=(UserOwnsPost() | UserIsAdmin()))
        def delete_post(user, post):
            post.delete()
        ```
        """
        if self.callback is None:
            raise NotImplemented('override in subclass')
        return self.callback(request, *args, **kwargs)


class CompositeGuard(Guard):
    """
    A CompositeGuard represents a boolean expression involving one or
    more Guard, which can themselves be other CompositeGuard. This
    subclass is used to form logical predicates involving multiple
    Guards.
    """

    def __init__(self, op: Text, lhs: Guard, rhs: Guard):
        super().__init__()
        self._op = op
        self._lhs = lhs
        self._rhs = rhs

    def __call__(self, request: 'Request', arguments: Dict) -> bool:
        return self.execute(request, arguments)

    @property
    def op_code(self):
        return self._op

    @property
    def lhs(self):
        return self._lhs

    @property
    def rhs(self):
        return self._rhs

    @property
    def description(self):
        opc = self._op
        if self._op == OP_CODE.NOT:
            return f'{opc} ({self._lhs.description})'
        if self._op == OP_CODE.AND:
            return f'({self._lhs.description} {opc} {self._rhs.description})'
        if self._op == OP_CODE.OR:
            return f'({self._lhs.description} {opc} {self._rhs.description})'

    def execute(self, request: 'Request', arguments: Dict):
        """
        Compute the boolean value of one or more nested Guard in a
        depth-first manner.
        """
        # compute LHS for both & and |.
        lhs_is_ok = self._lhs(request, arguments)

        if self._op == OP_CODE.AND:
            # We only need to check RHS if LHS isn't already False.
            if lhs_is_ok is False:
                raise GuardFailure(self._lhs)
            rhs_is_ok = self._rhs(request, arguments)
            if rhs_is_ok is False:
                raise GuardFailure(self._rhs)
        elif self._op == OP_CODE.OR:
            if lhs_is_ok is not False:
                rhs_is_ok = self._rhs(request, arguments)
                if rhs_is_ok is False:
                    raise GuardFailure(self)
        elif self._op == OP_CODE.NOT:
            if lhs_is_ok is not False:
                raise GuardFailure(self)

        return True



import sys

from typing import Text, List, Dict, Type

from google.protobuf.message import Message

from ravel.exceptions import RavelError
from ravel.util.misc_functions import get_class_name
from ravel.schema import Schema, fields

from ..util import get_stripped_schema_name
from .field_adapters import (
    FieldAdapter,
    SchemaFieldAdapter,
    NestedFieldAdapter,
    ScalarFieldAdapter,
    ArrayFieldAdapter,
)

class MessageGenerator(object):
    def __init__(self, adapters: Dict[Type[fields.Field], FieldAdapter]=None):
        # default field adapters indexed by appyratus schema Field types
        self.adapters = {
            Schema: SchemaFieldAdapter(),
            fields.Nested: NestedFieldAdapter(),
            fields.String: ScalarFieldAdapter('string'),
            fields.FormatString: ScalarFieldAdapter('string'),
            fields.Field: ScalarFieldAdapter('string'),
            fields.Email: ScalarFieldAdapter('string'),
            fields.Uuid: ScalarFieldAdapter('string'),
            fields.Bool: ScalarFieldAdapter('bool'),
            fields.Float: ScalarFieldAdapter('double'),
            fields.Int: ScalarFieldAdapter('uint64'),
            fields.DateTime: ScalarFieldAdapter('uint64'),
            fields.Dict: ScalarFieldAdapter('string'),
            fields.List: ArrayFieldAdapter(),
            fields.Set: ArrayFieldAdapter(),
            # XXX redundant to List?  does not exist in schema.fields
            #fields.Array: ArrayFieldAdapter(),
            # XXX do we add?  does not exist in schema.fields
            #fields.Enum: EnumFieldAdapter(),
            # XXX do we add?  does not exist in schema.fields
            #fields.Regexp: ScalarFieldAdapter('string'),
        }
        # upsert into default adapters dict from the `adapters` kwarg
        for field_type, adapter in (adapters or {}).items():
            self.adapters[field_type] = adapter
        # associate the generator with each adapter.
        for adapter in self.adapters.values():
            adapter.bind(self)

    def get_adapter(self, field) -> FieldAdapter:
        if isinstance(field, Schema):
            adapter = self.adapters[Schema]
        else:
            adapter = self.adapters.get(type(field))
            if adapter is None:
                unknown_field_type = type(field)
                for known_field_type, adapter in self.adapters.items():
                    if issubclass(unknown_field_type, known_field_type):
                        break
            if adapter is None:
                raise Exception(f'adapter for field {field} not found')
        return adapter

    def emit_resource_message(self, resource_type) -> Text:
        schema = resource_type.ravel.schema
        resolvers = resource_type.ravel.resolvers

        all_field_numbers = set(
            range(1, 1 + len(resolvers))
        )
        unavailable_field_numbers = {
            f.meta['field_no'] for f in schema.fields.values()
            if f.meta.get('field_no')
        }
        available_field_numbers = sorted(
            all_field_numbers - unavailable_field_numbers,
            reverse=True  # we want to pop in asc order below
        )

        body = []

        for resolver in resolvers.sort():
            line = None
            if resolver.name in schema.fields:
                adapter = self.get_adapter(resolver.field)
                if adapter is not None:
                    field = resolver.field
                    field_no = field.meta.get('field_no')
                    if not field_no:
                        field_no = available_field_numbers.pop()
                    line = adapter.emit(field=field, field_no=field_no)
            elif resolver.target is not None:
                adapter = self.get_adapter(resolver.schema)
                field_no = available_field_numbers.pop()
                line = adapter.emit(
                    field=resolver.schema,
                    field_no=field_no,
                    is_repeated=resolver.many
                )

            if line is not None:
                if line.strip().startswith('Customer created_at'):
                    import ipdb; ipdb.set_trace()
                body.append((field_no, f'   {line};\n'))

        header = f'message {get_class_name(resource_type)} {{\n'
        body = ''.join(x[1] for x in sorted(body))
        coda = '}'

        message = header + body + coda
        return message

    def emit(
        self,
        app: 'Application',
        schema_type: Type['Schema'],
        type_name: Text = None,
        depth=1,
        force=False,
    ) -> Text:
        """
        Recursively generate a protocol buffer message type declaration
        string from a given Schema class.
        """
        if isinstance(schema_type, (type, Schema)):
            type_name = get_stripped_schema_name(
                type_name or get_class_name(schema_type)
            )
        else:
            raise ValueError(
                'unrecognized schema type: "{}"'.format(schema_type)
            )

        # we don't want to create needless copies of to Resource schemas
        # while recursively generating nested message types.
        if (not force) and (type_name in app.manifest.resource_classes):
            return None

        prepared_data = []
        field_decls = []

        for f in schema_type.fields.values():
            # compute the "field number"
            field_no = f.meta.get('field_no', sys.maxsize)
            if field_no is None:
                raise ValueError(f'f has no protobuf field number')

            # get the field adapter
            adapter = self.get_adapter(f)
            if not adapter:
                raise Exception('no adapter for type {}'.format(type(f)))

            # store in intermediate data structure for purpose of sorting by
            # field numbers
            prepared_data.append((field_no, f, adapter))

        # emit field declarations in order of field number ASC
        sorted_data = sorted(prepared_data, key=lambda x: x[0])
        for (field_no, field, adapter) in sorted_data:
            field_decl = adapter.emit(field, field_no)
            field_decls.append(('  ' * depth) + field_decl + ';')

        nested_message_types = []
        for nested_schema in {type(s) for s in schema_type.children}:
            type_source = self.emit(
                app, nested_schema, depth=depth + 1, force=force
            )
            if type_source is not None:
                nested_message_types.append(type_source)

        MESSAGE_TYPE_FSTR = (
            (('   ' * (depth - 1)) + '''message {type_name} ''') + \
            ('''{{\n{nested_message_types}{field_lines}\n''') + \
            (('   ' * (depth - 1)) + '''}}''')
        )

        # emit the message declaration "message Foo { ... }"
        return MESSAGE_TYPE_FSTR.format(
            type_name=type_name,
            nested_message_types=(
                '\n'.join(nested_message_types) + '\n'
                if nested_message_types else ''
            ),
            field_lines='\n'.join(field_decls),
        ).rstrip()



import falcon

from typing import Dict, Tuple, Type, Text, List
from inspect import Parameter

from ravel.app.middleware import Middleware
from ravel.apps.web import Endpoint
from ravel.app.base import ActionDecorator
from ravel.ext.falcon.service import FalconService
from ravel.ext.falcon.constants import HTTP_METHODS, HTTP_OPTIONS

DEFAULT_ALLOW_ORIGIN = '*'
DEFAULT_ALLOW_METHODS = HTTP_METHODS
DEFAULT_ALLOW_HEADERS = (
    'Authorization',
    'Content-Type',
    'Accept',
    'Origin',
    'User-Agent',
    'DNT',
    'Cache-Control',
    'X-Mx-ReqToken',
    'Keep-Alive',
    'X-Requested-With',
    'If-Modified-Since',
    'Pragma',
    'Expires',
)


class SetHttpCorsResponseHeaders(Middleware):
    """
    # CORS Middleware

    This middleware sets various Access-Control headers needed for CORS.
    """

    def __init__(
        self,
        allow_origin: Text = DEFAULT_ALLOW_ORIGIN,
        allow_headers: List[Text] = DEFAULT_ALLOW_HEADERS,
        allow_methods: List[Text] = DEFAULT_ALLOW_METHODS,
    ):
        self._allow_origin = allow_origin
        self._allow_headers = allow_headers
        self._allow_methods = allow_methods
        self._cors_headers = {
            'Access-Control-Allow-Origin': self._allow_origin,
            'Access-Control-Allow-Methods': ','.join(self._allow_methods),
            'Access-Control-Allow-Headers': ','.join(self._allow_headers),
        }

    @property
    def app_types(self) -> Tuple[Type['Application']]:
        return (FalconService, )

    def on_bootstrap(self):
        """
        Ensure that every registered route has at least a dummy endpoint
        registered for the OPTIONS HTTP method, adding endpoints as needed.
        """
        for res in self.app.falcon_resources.values():
            if not res.is_method_supported(HTTP_OPTIONS):
                endpoint = Endpoint.from_function(
                    app=self.app,
                    func=lambda *args, **kwargs: None,
                    method=HTTP_OPTIONS,
                    route=res.route,
                )
                res.add_endpoint(endpoint)

    def pre_request(
        self,
        action: 'Action',
        request: 'Request',
        raw_args: Tuple,
        raw_kwargs: Dict
    ):
        """
        Set CORS headers on OPTIONS requests as well as tell Falcon and Ravel
        to abort both further middleware processing and the target/respondor
        method, going right to post-request middleware instead.
        """
        falcon_request, falcon_response = raw_args[:2]
        falcon_response.set_headers(self._cors_headers)


from random import randint

from typing import List, Set, Text, Dict

from ravel.util import is_batch
from ravel.util.loggers import console
from ravel.batch import Batch
from ravel.constants import ID, REV


class Executor(object):

    def __init__(self, simulate: bool = False):
        self.simulate = simulate

    def execute(self, query: 'Query', sources: List['Resource'] = None) -> 'Batch':
        # extract information needed to perform execution logic
        info = self._analyze_query(query)

        # fetch target fields from the DAL and execute all other requested
        # resolvers that don't correspond to fields, merging the results into
        # the fetched resources.
        resources = self._fetch_resources(query, info['fields'], sources)
        self._execute_requests(query, resources, info['requests'])
        return resources

    def _analyze_query(self, query) -> Dict:
        fields_to_fetch = {ID, REV}
        requests_to_execute = set()
        schema = query.target.ravel.schema
        for request in query.requests.values():
            field = schema.fields.get(request.resolver.name)
            if field and not field.meta.get('ravel_on_resolve'):
                fields_to_fetch.add(field.name)
            else:
                requests_to_execute.add(request)
        return {
            'fields': fields_to_fetch,
            'requests': requests_to_execute,
        }

    def _fetch_resources(self, query: 'Query', fields: Set[Text], sources: List['Resource'] = None) -> List['Resource']:
        """
        # Fetch Resources
        This is where we take the query params and send them to the store
        """
        resource_type = query.target
        predicate = query.parameters.where
        if query.target.ravel.app is None:
            import ipdb
            ipdb.set_trace()
            print('=' * 100)
        mode = query.target.ravel.app.mode
        kwargs = query.parameters.to_dict()

        if predicate is None:
            predicate = resource_type._id != None

        if mode == 'normal' and (not self.simulate):
            store = resource_type.ravel.local.store
            records = store.query(predicate, fields=fields, **kwargs)
            batch = resource_type.Batch(resource_type(state=record).clean() for record in records)
        else:
            values = predicate.satisfy() if predicate else None
            count = kwargs.get('limit') or randint(1, 10)
            batch = resource_type.Batch.generate(resolvers=fields, values=values, count=count)
            if query.parameters.order_by:
                batch.sort(query.parameters.order_by)

        if sources:
            batch.extend(sources)

        return batch

    def _execute_requests(self, query, resources, requests):
        for request in requests:
            request.query = query
            resolver = request.resolver
            # first try resolving in batch
            if len(resources) > 1:
                results = resolver.resolve_batch(resources, request)
                if results:
                    for resource in resources:
                        value = results.get(resource)
                        if resolver.many and value is None:
                            value = self.target.Batch()
                        elif (not resolver.many) and is_batch(value):
                            value = None
                        resource.internal.state[resolver.name] = value
                    continue

            # default on resolving for each resource separatesly
            for resource in resources:
                value = resolver.resolve(resource, request)
                resource.internal.state[resolver.name] = value



import os
import glob

from typing import Text, List, Set, Dict, Tuple
from collections import defaultdict
from datetime import datetime
from threading import RLock

import yaml

from appyratus.env import Environment
from appyratus.files import BaseFile, Yaml
from appyratus.schema.fields import UuidString
from appyratus.utils.dict_utils import DictObject, DictUtils
from appyratus.utils.path_utils import PathUtils
from appyratus.utils.string_utils import StringUtils

from ravel.util.misc_functions import import_object
from ravel.util.loggers import console
from ravel.constants import ID, REV
from ravel.exceptions import RavelError

from .base import Store
from .simulation_store import SimulationStore


class StoreError(RavelError):
    pass


class MissingBootstrapParameterError(StoreError):
    error_code = 'missing-bootstrap-param'
    error_message = 'Missing Bootstrap Parameter `{id}`'
    error_help = "Add the missing parameter to your manifest's bootstrap settings"


class FilesystemStore(Store):

    env = Environment()
    root = None
    paths = None

    def __init__(
        self,
        ftype: Text = None,
        extension: Text = None,
    ):
        super().__init__()

        self._paths = DictObject()
        self._cache_store = SimulationStore()

        # convert the ftype string arg into a File class ref
        if not ftype:
            self._ftype = Yaml
        elif not isinstance(ftype, BaseFile):
            self._ftype = import_object(ftype)

        assert issubclass(self.ftype, BaseFile)

        known_extension = self._ftype.has_extension(extension)
        if known_extension:
            self._extension = known_extension
        if extension is None:
            self._extension = self._ftype.default_extension()

    @property
    def paths(self):
        return self._paths

    @property
    def extension(self):
        return self._extension

    @property
    def ftype(self):
        return self._ftype

    @classmethod
    def on_bootstrap(
        cls,
        ftype: Text = None,
        root: Text = None,
        use_recursive_merge=True,
        store_primitives=False,
        prefetch: bool = True,
        yaml_loader_class: Text = 'FullLoader'
    ):
        cls.ftype = import_object(ftype) if ftype else Yaml
        cls.root = PathUtils.expand_path(root or cls.root)
        cls.use_recursive_merge = use_recursive_merge
        cls.store_primitives = store_primitives
        cls.do_prefetch = prefetch

        if 'yaml' in cls.ftype.extensions():
            cls.yaml_loader_class = getattr(yaml, yaml_loader_class, None)
        else:
            cls.yaml_loader_class = None

        if not cls.root:
            raise MissingBootstrapParameterError('missing parameter: root')

    def on_bind(
        self,
        resource_type,
        root: Text = None,
        ftype: BaseFile = None,
        store_primitives=None,
        prefetch: bool = None,
        yaml_loader_class: Text = None,
    ):
        """
        Ensure the data dir exists for this Resource type.
        """
        if isinstance(ftype, str):
            self.ftype = import_object(ftype)

        if store_primitives is not None:
            self.store_primitives = store_primitives

        if prefetch is not None:
            self.do_prefetch = prefetch

        if yaml_loader_class is not None:
            if self.ftype.lower() == 'yaml':
                self.yaml_loader_class = getattr(yaml, yaml_loader_class, None)
            else:
                self.yaml_loader_class = None

        self.paths.root = root or self.root
        self.paths.records = os.path.join(self.paths.root, StringUtils.snake(resource_type.__name__))

        os.makedirs(self.paths.records, exist_ok=True)

        # bootstrap, bind, and backfill the in-memory cache
        if self.do_prefetch:
            self.bust_cache(self.do_prefetch)

    def bust_cache(self, prefetch=False):
        self._cache_store = SimulationStore()
        if not self._cache_store.is_bootstrapped():
            self._cache_store.bootstrap(self.resource_type.ravel.app)

        self._cache_store.bind(self.resource_type)

        if prefetch:
            self._cache_store.create_many(record for record in self.fetch_all(ignore_cache=True).values() if record)

    @classmethod
    def has_transaction(cls):
        return False

    @classmethod
    def begin(cls, **kwargs):
        pass

    @classmethod
    def commit(cls, **kwargs):
        pass

    @classmethod
    def rollback(cls, **kwargs):
        pass

    def create_id(self, record):
        return record.get(ID, UuidString.next_id())

    def exists(self, _id: Text) -> bool:
        return BaseFile.exists(self.mkpath(_id))

    def exists_many(self, _ids: Set) -> Dict[object, bool]:
        return {_id: self.exists(_id) for _id in _ids}

    def create(self, record: Dict) -> Dict:
        _id = self.create_id(record)
        record = self.update(_id, record)
        record[ID] = _id
        self._cache_store.create(record)
        return record

    def create_many(self, records):
        created_records = []
        for record in records:
            created_records.append(self.create(record))
        self._cache_store.create_many(created_records)
        return created_records

    def count(self) -> int:
        fnames = glob.glob(f'{self.paths.records}/*.{self.extension}')
        return len(fnames)

    def fetch(self, _id, fields=None) -> Dict:
        records = self.fetch_many([_id], fields=fields)
        record = records.get(_id) if records else {}
        return record

    def fetch_many(self, _ids: List = None, fields: Set[Text] = None, ignore_cache=False) -> Dict:
        """
        """
        if not _ids:
            _ids = []

        # reduce _ids to its unique members by making it a set
        if not isinstance(_ids, set):
            all_ids = set(_ids)
        else:
            all_ids = _ids

        ids_to_fetch_from_fs = set()

        # we do not want to ignore the cache here
        if not ignore_cache:
            cached_records = self._cache_store.fetch_many(all_ids, fields=fields)
            for record_id, record in cached_records.items():
                if record is None:
                    ids_to_fetch_from_fs.add(record_id)
        # otherwise we will go straight to the filesystem
        else:
            cached_records = {}
            ids_to_fetch_from_fs = self._fetch_all_ids()

        # if there are any remaining ID's not returned from cache,
        # fetch them from file system
        if ids_to_fetch_from_fs:

            # prepare the set of field names to fetch
            fields = fields if isinstance(fields, set) else set(fields or [])
            if not fields:
                fields = set(self.resource_type.Schema.fields.keys())
            fields |= {ID, REV}

            records = {}
            non_null_records = []

            for _id in ids_to_fetch_from_fs:
                fpath = self.mkpath(_id)
                try:
                    record = self.ftype.read(fpath, loader_class=self.yaml_loader_class)
                except FileNotFoundError:
                    records[_id] = None
                    console.debug(message='file not found by filesystem store', data={'filepath': fpath})
                    continue

                if record:
                    record, errors = self.schema.process(record)
                    if errors:
                        raise Exception(f'validation error while loading ' f'{_id}.{self.extension}')
                    record.setdefault(ID, _id)
                    records[_id] = {k: record.get(k) for k in fields}

                    non_null_records.append(record)

                    # if for some reason a file was created manually
                    # with a _rev, we create one here and save it
                    if REV not in record:
                        record[REV] = self.increment_rev()
                        self.update(_id, record)
                else:
                    records[ID] = None

            self._cache_store.create_many(non_null_records)
            cached_records.update(records)

        return cached_records

    def fetch_all(self, fields: Set[Text] = None, ignore_cache=False) -> Dict:
        return self.fetch_many(None, fields=fields, ignore_cache=ignore_cache)

    def update(self, _id, data: Dict) -> Dict:
        fpath = self.mkpath(_id)
        base_record = self.ftype.read(fpath, loader_class=self.yaml_loader_class)

        schema = self.resource_type.ravel.schema
        base_record, errors = schema.process(base_record)

        if base_record:
            # this is an upsert
            if self.use_recursive_merge:
                record = DictUtils.merge(base_record, data)
            else:
                record = dict(base_record, **data)
        else:
            record = data

        record[REV] = self.increment_rev(record.get(REV))
        if ID not in record:
            record[ID] = _id

        if self.store_primitives:
            json = self.app.json
            self.ftype.write(path=fpath, data=json.decode(json.encode(record)))
        else:
            self.ftype.write(path=fpath, data=record)

        self._cache_store.update(_id, record)
        return record

    def update_many(self, _ids: List, updates: List = None) -> Dict:
        return {_id: self.update(_id, data) for _id, data in zip(_ids, updates)}

    def delete(self, _id) -> None:
        self._cache_store.delete(_id)
        fpath = self.mkpath(_id)
        os.remove(fpath)

    def delete_many(self, _ids: List) -> None:
        for _id in _ids:
            self.delete(_id)

    def delete_all(self):
        _ids = self._fetch_all_ids()
        self.delete_many(_ids)

    def query(self, *args, **kwargs):
        if self._cache_store is not None and not self._cache_store.count():
            self.fetch_all(ignore_cache=True)

        return self._cache_store.query(*args, **kwargs)

    def mkpath(self, fname: Text) -> Text:
        fname = self.ftype.format_file_name(fname)
        return os.path.join(self.paths.records, fname)

    def _fetch_all_ids(self):
        _ids = set()
        for fpath in glob.glob(f'{self.paths.records}/*.{self.extension}'):
            fname = fpath.split('/')[-1]
            basename = os.path.splitext(fname)[0]
            _ids.add(basename)
        return _ids



import inspect
import socket

from uuid import UUID
from importlib import import_module
from types import GeneratorType
from typing import (
    List, Dict, ForwardRef, Text, Tuple, Set, Type,
    _GenericAlias as GenericAlias, Callable
)

from copy import deepcopy


DictKeySet = type({}.keys())
DictValueSet = type({}.values())


def is_sequence(obj) -> bool:
    """
    Return True if obj is a generic sequence type, like a list or tuple.
    """
    return isinstance(obj, (list, tuple, set, DictKeySet, DictValueSet))


def import_object(dotted_path: Text) -> object:
    """
    Import an object from a module, given a dotted path to it.
    """
    obj_path = dotted_path.split('.')

    if len(obj_path) < 2:
        raise ImportError(dotted_path)

    module_path_str = '.'.join(obj_path[:-1])
    obj_name = obj_path[-1]

    try:
        module = import_module(module_path_str)
        obj = getattr(module, obj_name)
    except Exception:
        raise ImportError(f'failed to import {dotted_path}')

    return obj


def get_class_name(obj):
    if not obj:
        return None
    if isinstance(obj, type):
        return obj.__name__
    else:
        return obj.__class__.__name__


def get_callable_name(obj):
    if inspect.isfunction(obj):
        return obj.__name__
    elif inspect.ismethod(obj):
        return obj.__func__.__name__
    else:
        raise ValueError(str(obj))


def remove_keys(
    records: List[Dict], keys: Set[Text], in_place=True
) -> List[Dict]:
    """
    Remove keys from a dict. Non-recursive.
    """
    records_out = []
    for record in records:
        if record is not None:
            if not in_place:
                record = deepcopy(record)
            for k in keys:
                record.pop(k, None)
        records_out.append(record)
    return records_out


def normalize_to_tuple(obj):
    """
    If obj is a tuple, return it as-is; otherwise, return it in a tuple with it
    as its single element.
    """
    if obj is not None:
        if isinstance(obj, (list, tuple)):
            return tuple(obj)
        return (obj, )
    return tuple()


def extract_res_info_from_annotation(annotation) -> Tuple[bool, Text]:
    """
    Return a tuple of metadata pertaining to `obj`, which is some object
    used in a type annotation, passed in by the caller.
    """
    key = None
    many = False

    if isinstance(annotation, str):
        key = annotation.split('.')[-1]
    elif isinstance(annotation, type):
        key = annotation.__name__.split('.')[-1]
    elif isinstance(annotation, ForwardRef):
        key = annotation.__forward_arg__
    elif (
        (isinstance(annotation, GenericAlias)) and
        (annotation._name in {'List', 'Tuple', 'Set'})
    ):
        if annotation.__args__:
            arg = annotation.__args__[0]
            key = extract_res_info_from_annotation(arg)[1]
            many = True

    if key == '_empty':
        key = None

    return (many, key)


def flatten_sequence(seq) -> List:
    flattened = []
    if not seq:
        return flattened
    for obj in seq:
        if is_sequence(obj):
            flattened.extend(flatten_sequence(obj))
        elif isinstance(obj, GeneratorType):
            flattened.extend(list(obj))
        elif isinstance(obj, dict):
            flattened.extend(obj.values())
        else:
            flattened.append(obj)
    return flattened


def inject(func: Callable, data: Dict):
    """
    Dynamically inject data dict into a function's lexcial scope.
    """
    func.__globals__.update(data)


def union(sequences):
    if sequences:
        if len(sequences) == 1:
            return sequences[0]
        else:
            return set.union(*sequences)
    else:
        return set()


def is_port_in_use(addr: Text) -> bool:
    """
    Utility method for determining if the server address is already in use.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        host, port_str = addr.split(':')
        sock.bind((host, int(port_str)))
        return False
    except OSError as err:
        if err.errno == 48:
            return True
        else:
            raise err
    finally:
        sock.close()





from .application_router import ApplicationRouter
from .type_checking import (
    is_resource,
    is_batch,
    is_resource_type,
    is_batch_type,
)
from .misc_functions import (
    get_class_name,
    is_sequence,
    flatten_sequence,
    is_port_in_use,
    union,
)
from .uuid_util import random_uuid


from typing import Text

from appyratus.utils.time_utils import TimeUtils

from ravel.resource import Resource
from ravel.schema import DateTime, String


class GraphqlQuery(Resource):
    created_at = DateTime(default=TimeUtils.utc_now, required=True, private=True)
    source = String()

    @classmethod
    def __abstract__(cls):
        return True

    @classmethod
    def on_bootstrap(cls, *args, **kwargs):
        from .graphql_interpreter import GraphqlInterpreter

        cls._interpreter = GraphqlInterpreter(cls)

    @classmethod
    def interpret(cls, query: Text) -> 'Query':
        """
        Transform the GraphQL query string into a Ravel query.
        """
        return cls._interpreter.interpret(query)



from typing import Dict, List, Text, Type

from ravel.app.base import Application, Action


class Repl(Application):
    """
    Repl is a Application that collects all registered functions and
    injects them into an interactive Python shell, or REPL. This is useful for
    experimenting with an API from a command-line interface.
    """

    def __init__(self, autoreload=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._autoreload = autoreload
        self.shell = None

    @property
    def action_type(self) -> Type['ReplFunction']:
        return ReplFunction

    @property
    def autoreload(self) -> bool:
        return self._autoreload

    @autoreload.setter
    def autoreload(self, new_value: bool):
        if (not new_value) and self._autoreload:
            self.shell.extension_manager.unload_extension('autoreload')
        elif new_value and (not self._autoreload):
            self.shell.extension_manager.load_extension('autoreload')
        self._autoreload = new_value

    @property
    def namespace(self) -> Dict:
        """
        iPython's embedded shell session namespace dict. Update this dict from
        methods when you want to, say, rerun and reload fixture data inside a
        REPL, like:

        ```python3
            @repl()
            def reset_fixtures():
                fixtures = {
                    'foo': Foo().save(),
                    'bar': Bar().save()
                }
                repl.namespace.update(fixtures)
                return fixtures
            ```

        Now, inside the REPL session, you can do `reset_fixtures()` to reset the
        global variables available to you in the shell.
        """
        if not self.is_bootstrapped:
            return super().namespace
        else:
            return self.shell.user_ns

    @property
    def functions(self) -> List[Text]:
        """
        Get list of names of all registered functions in the REPL.
        """
        return sorted(self.actions.keys())

    def on_start(self):
        """
        Start a new REPL with all registered functions available in the REPL
        namespace.
        """
        from IPython.terminal.embed import InteractiveShellEmbed
        
        self.shell = InteractiveShellEmbed()

        if self._autoreload:
            self.shell.extension_manager.load_extension('autoreload')

        # build the shell namespace
        local_ns = {}
        local_ns['repl'] = self

        local_ns.update(self._namespace)
        local_ns.update(self.actions)
        local_ns.update(self.manifest.resource_classes)
        local_ns.update(self.manifest.store_classes)

        # enter an ipython shell
        self.shell.mainloop(local_ns=local_ns)

    def on_response(
        self,
        repl_function: 'ReplFunction',
        result,
        *args,
        **kwargs
    ) -> object:
        if repl_function.memoized:
            repl_function.memoize(result)
        return super().on_response(repl_function, result, *args, **kwargs)


class ReplFunction(Action):
    def __init__(self, func, decorator):
        super().__init__(func, decorator)
        self.return_values = []

    @property
    def memoized(self) -> bool:
        return self.decorator.kwargs.get('memoized', False)

    def memoize(self, value):
        self.return_values.append(value)



import inspect

from pprint import pprint
from ravel.util.type_checking import is_resource_type
from typing import List
from appyratus.cli import (
    CliProgram,
    OptionalArg,
    PositionalArg,
    ListArg,
    FlagArg,
    FileArg,
    Subparser,
)
from appyratus.files import Yaml
from appyratus.utils.string_utils import StringUtils
from appyratus.utils.sys_utils import SysUtils

from ravel.util import is_batch, is_resource, is_resource_type
from ravel.app.base import Application, ActionDecorator, Action


class CliApplication(Application):
    """
    This Application subclass will create a CliProgram (command-line
    interace) out of the registered functions.
    """

    def __init__(
        self,
        name=None,
        version=None,
        tagline=None,
        defaults=None,
        echo=False,
        debug_level=2,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._commands = []
        self._echo = echo
        self._debug_level = debug_level
        self._cli_program = None
        self._cli_args = None
        self._cli_program_kwargs = {
            'name': name,
            'version': version,
            'tagline': tagline,
            'defaults': defaults,
        }

    @property
    def action_type(self):
        return CliCommand

    def on_bootstrap(self, cli_args=None):
        """
        Collect subparsers and build cli program
        """
        self._cli_args = cli_args

        subparsers = [
            c.subparser for c in self.actions.values() if c.subparser
        ]

        self._cli_program = CliProgram(
            subparsers=subparsers,
            cli_args=self._cli_args,
            custom_dtype_converter=(
                lambda x: str if is_resource_type(x) else None
            ),
            **self._cli_program_kwargs
        )

    def on_start(self, debug=True):
        """
        Run the CliProgram.
        """
        if debug:
            return SysUtils.safe_main(
                self._cli_program.run, debug_level=self._debug_level
            )
        else:
            return self._cli_program.run()

    def on_request(self, action, *args, **kwargs):
        """
        Extract command line arguments and bind them to the arguments expected
        by the registered function's signature.
        """
        args, kwargs = [], {}

        # Python's ArgumentParser Arguments object, with the addition of a
        # `unknown` attribute, containing all raw keyword CLI arguments not
        # explicitly defined in the parser:
        cli_args = self._cli_program.cli_args

        for idx, (k, param) in enumerate(action.signature.parameters.items()):
            if not idx:
                continue
            value = getattr(cli_args, k, None)
            if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                if param.default is inspect._empty:
                    args.append(value)
                else:
                    kwargs[k] = value
            elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
                args.append(value)
            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                kwargs[k] = value

        unknown = getattr(cli_args, 'unknown', {})
        kwargs.update(unknown)

        return (args, kwargs)

    def on_response(self, action, result, *args, **kwargs):
        response = super().on_response(action, result, *args, **kwargs)
        dumped_result = _dump_result_obj(response)
        if self._echo:
            output_format = getattr(self._cli_program.cli_args, 'format', None)
            formatted_result = _format_result_data(dumped_result, output_format)
            if isinstance(formatted_result, str):
                print(formatted_result)
            else:
                pprint(formatted_result)
        return dumped_result


def _format_result_data(data, output_format):
    if output_format == 'yaml':
        return Yaml.from_data(data)
    else:
        return data


def _dump_result_obj(obj):
    if is_resource(obj) or is_batch(obj):
        return obj.dump()
    elif isinstance(obj, (list, set, tuple)):
        return [_dump_result_obj(x) for x in obj]
    elif isinstance(obj, dict):
        return {k: _dump_result_obj(v) for k, v in obj.items()}
    else:
        return obj


class CliCommand(Action):
    """
    CliCommand represents a top-level CliProgram Subparser.
    """

    def __init__(self, func, decorator):
        super().__init__(func, decorator)
        self.program_name = None
        self.subparser_kwargs = None
        self.subparser_type = None
        self.subparser = None
        self.schema = decorator.kwargs.get('schema')

    def on_bootstrap(self):
        self.program_name = self.decorator.kwargs.get('name', self.name)
        self.subparser_kwargs = self._build_subparser_kwargs(self.exc)
        self.subparser_type = self.decorator.kwargs.get('subparser', Subparser)
        self.subparser = self.subparser_type(**self.subparser_kwargs)

    def _build_subparser_kwargs(self, func):
        parser_kwargs = self.decorator.kwargs.get('parser') or {}
        custom_args = self.decorator.kwargs.get('args') or []
        cli_args = self._build_cli_args(func, custom_args)
        name = StringUtils.dash(parser_kwargs.get('name') or self.program_name)
        return dict(parser_kwargs, **{
            'name': name,
            'args': cli_args,
            'perform': self,
        })

    def _build_cli_args(self, func, custom_args: List = None):
        if custom_args is None:
            custom_args = []

        args = []
        custom_args_by_name = {a.name: a for a in custom_args}
        for k, param in self.signature.parameters.items():
            if k == 'request':
                # we do not want to include `request` in with the supported
                # args for passing into the action
                # TODO is this the best way?
                continue
            if k in custom_args_by_name.keys():
                # if a custom argument was passed into the decorator, then this
                # should take precedent over the inferred arg created in this
                # iteration
                args.append(custom_args_by_name[k])
                continue

            arg = None
            arg_type = None
            dtype = param.annotation

            if param.annotation is inspect._empty:
                dtype = None

            if is_resource_type(dtype):
                dtype = str

            arg_params = {
                'name': k,
                'dtype': dtype,
            }

            if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                if param.default is inspect._empty:
                    arg_type = PositionalArg
                else:
                    arg_type = OptionalArg
                    arg_params['default'] = param.default
            elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
                arg_type = PositionalArg
            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                arg_type = OptionalArg
                arg_params['default'] = param.default
            if 'List' in str(dtype):
                arg_type = ListArg
            elif 'bool' in str(dtype):
                arg_type = FlagArg
            elif 'File' in str(dtype):
                arg_type = FileArg

            if arg_type:
                arg = arg_type(**arg_params)
                # flag args are processed in the same way as everything else,
                # which means that their null value for arg.default throws things
                # off, so we hackily set it here on the arg object.
                if arg_params.get('default') and (not arg.default):
                    arg.default = True
            if arg is not None:
                args.append(arg)
        return args


Cli = CliApplication


import re


HTTP_GET = 'GET'
HTTP_POST = 'POST'
HTTP_PUT = 'PUT'
HTTP_PATCH = 'PATCH'
HTTP_DELETE = 'DELETE'
HTTP_OPTIONS = 'OPTIONS'

HTTP_METHODS = frozenset({
    HTTP_GET, HTTP_POST, HTTP_PUT, HTTP_PATCH, HTTP_DELETE, HTTP_OPTIONS
})

RE_HANDLER_METHOD = re.compile(r'^on_([a-z]+)$')



from typing import Text, List, Dict
from appyratus.utils.template_utils import JinjaTemplateEnvironment

from .renderer import TemplateRenderer


class JinjaTemplateRenderer(TemplateRenderer):

    def __init__(self, path: List[Text], **jinja_env_kwargs):
        search_path = path if isinstance(path, str) else [path]
        self._jinja_env = JinjaTemplateEnvironment(search_path=path, **jinja_env_kwargs)

    def render(self, template: Text, context: Dict = None) -> Text:
        template_obj = self._jinja_env.from_filename(template)
        return template_obj.render(context or {})
